from __future__ import annotations

import time
from pathlib import Path
from typing import Any

import numpy as np
import torch
from torch import nn
from torch.utils.data import DataLoader, Dataset

from .ae import CircularConv1d
from .ks import params_from_config, power_spectrum
from .ks_data import _spectrum_rel_l1, generate_autoencoder_data
from .utils import best_device, ensure_dir, format_seconds, save_json, set_seed


class KSDataset(Dataset):
    def __init__(self, states: np.ndarray, mean: float, std: float, augment_shift: bool) -> None:
        self.states = ((states.astype(np.float32) - float(mean)) / float(std)).astype(np.float32)
        self.augment_shift = bool(augment_shift)
        self.n = int(states.shape[-1])

    def __len__(self) -> int:
        return int(self.states.shape[0])

    def __getitem__(self, idx: int) -> torch.Tensor:
        x = torch.from_numpy(self.states[idx]).unsqueeze(0)
        if self.augment_shift:
            shift = int(torch.randint(0, self.n, ()).item())
            x = torch.roll(x, shifts=shift, dims=-1)
        return x


class KSConvAutoencoder(nn.Module):
    def __init__(
        self,
        n: int = 1024,
        latent_channels: int = 2,
        kz: int = 64,
        base_channels: int = 32,
    ) -> None:
        super().__init__()
        if n % kz != 0:
            raise ValueError("N must be divisible by Kz.")
        downsample = n // kz
        if downsample <= 0 or downsample & (downsample - 1):
            raise ValueError("N / Kz must be a power of two.")
        self.n = int(n)
        self.latent_channels = int(latent_channels)
        self.kz = int(kz)
        self.base_channels = int(base_channels)
        stages = int(np.log2(downsample))
        b = int(base_channels)
        stage_channels = [min(b * (2 ** min(stage, 2)), 4 * b) for stage in range(stages)]

        encoder: list[nn.Module] = [CircularConv1d(1, b, 7), nn.GELU()]
        in_ch = b
        for out_ch in stage_channels:
            encoder.extend(
                [
                    CircularConv1d(in_ch, out_ch, 5, stride=2),
                    nn.GELU(),
                    CircularConv1d(out_ch, out_ch, 5),
                    nn.GELU(),
                ]
            )
            in_ch = out_ch
        encoder.extend(
            [
                CircularConv1d(in_ch, 2 * b, 3),
                nn.GELU(),
                CircularConv1d(2 * b, self.latent_channels, 3),
            ]
        )
        self.encoder = nn.Sequential(*encoder)

        hidden = stage_channels[-1] if stage_channels else b
        decoder: list[nn.Module] = [CircularConv1d(self.latent_channels, hidden, 3), nn.GELU()]
        in_ch = hidden
        for out_ch in reversed(stage_channels):
            decoder.extend(
                [
                    nn.Upsample(scale_factor=2, mode="nearest"),
                    CircularConv1d(in_ch, out_ch, 5),
                    nn.GELU(),
                    CircularConv1d(out_ch, out_ch, 5),
                    nn.GELU(),
                ]
            )
            in_ch = out_ch
        decoder.extend([CircularConv1d(in_ch, b, 5), nn.GELU(), CircularConv1d(b, 1, 3)])
        self.decoder = nn.Sequential(*decoder)

    def encode(self, x: torch.Tensor) -> torch.Tensor:
        return self.encoder(x)

    def decode(self, z: torch.Tensor) -> torch.Tensor:
        return self.decoder(z)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.decode(self.encode(x))


class KSGlobalAutoencoder(nn.Module):
    def __init__(
        self,
        n: int = 1024,
        latent_dim: int = 256,
        bottleneck_grid: int = 64,
        base_channels: int = 32,
    ) -> None:
        super().__init__()
        if n % bottleneck_grid != 0:
            raise ValueError("N must be divisible by global bottleneck_grid.")
        downsample = n // bottleneck_grid
        if downsample <= 0 or downsample & (downsample - 1):
            raise ValueError("N / bottleneck_grid must be a power of two.")
        self.n = int(n)
        self.latent_dim = int(latent_dim)
        self.bottleneck_grid = int(bottleneck_grid)
        self.base_channels = int(base_channels)
        stages = int(np.log2(downsample))
        b = int(base_channels)
        stage_channels = [min(b * (2 ** min(stage, 2)), 4 * b) for stage in range(stages)]

        encoder: list[nn.Module] = [CircularConv1d(1, b, 7), nn.GELU()]
        in_ch = b
        for out_ch in stage_channels:
            encoder.extend(
                [
                    CircularConv1d(in_ch, out_ch, 5, stride=2),
                    nn.GELU(),
                    CircularConv1d(out_ch, out_ch, 5),
                    nn.GELU(),
                ]
            )
            in_ch = out_ch
        hidden = stage_channels[-1] if stage_channels else b
        encoder.extend([CircularConv1d(in_ch, hidden, 3), nn.GELU()])
        self.encoder_conv = nn.Sequential(*encoder)
        self.encoder_dense = nn.Linear(hidden * self.bottleneck_grid, self.latent_dim)

        self.decoder_dense = nn.Linear(self.latent_dim, hidden * self.bottleneck_grid)
        decoder: list[nn.Module] = [CircularConv1d(hidden, hidden, 3), nn.GELU()]
        in_ch = hidden
        for out_ch in reversed(stage_channels):
            decoder.extend(
                [
                    nn.Upsample(scale_factor=2, mode="nearest"),
                    CircularConv1d(in_ch, out_ch, 5),
                    nn.GELU(),
                    CircularConv1d(out_ch, out_ch, 5),
                    nn.GELU(),
                ]
            )
            in_ch = out_ch
        decoder.extend([CircularConv1d(in_ch, b, 5), nn.GELU(), CircularConv1d(b, 1, 3)])
        self.decoder_conv = nn.Sequential(*decoder)

    def encode(self, x: torch.Tensor) -> torch.Tensor:
        h = self.encoder_conv(x)
        return self.encoder_dense(h.flatten(start_dim=1))

    def decode(self, z: torch.Tensor) -> torch.Tensor:
        h = self.decoder_dense(z)
        h = h.reshape(z.shape[0], -1, self.bottleneck_grid)
        return self.decoder_conv(h)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.decode(self.encode(x))


class KSGlobalPCAAutoencoder(nn.Module):
    def __init__(self, n: int = 1024, latent_dim: int = 256) -> None:
        super().__init__()
        self.n = int(n)
        self.latent_dim = int(latent_dim)
        self.bottleneck_grid = 0
        self.base_channels = 0
        self.register_buffer("pca_mean", torch.zeros(self.n, dtype=torch.float32))
        self.register_buffer("components", torch.zeros(self.latent_dim, self.n, dtype=torch.float32))

    def encode(self, x: torch.Tensor) -> torch.Tensor:
        flat = x.squeeze(1)
        return (flat - self.pca_mean) @ self.components.T

    def decode(self, z: torch.Tensor) -> torch.Tensor:
        flat = z @ self.components + self.pca_mean
        return flat.unsqueeze(1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.decode(self.encode(x))


def checkpoint_path(config: dict[str, Any]) -> Path:
    model = config["model"]
    ae = config["ae"]
    length = float(model["length"])
    length_label = int(length) if float(length).is_integer() else str(length).replace(".", "p")
    return Path(config["paths"]["model_dir"]) / (
        f"ks_ae_N{int(model['N'])}_L{length_label}_C{int(ae['latent_channels'])}_Kz{int(ae['Kz'])}.pt"
    )


def global_checkpoint_path(config: dict[str, Any]) -> Path:
    model = config["model"]
    ae = config["global_ae"]
    length = float(model["length"])
    length_label = int(length) if float(length).is_integer() else str(length).replace(".", "p")
    return Path(config["paths"]["model_dir"]) / (
        f"ks_global_ae_N{int(model['N'])}_L{length_label}_Z{int(ae['latent_dim'])}.pt"
    )


def _metrics_path(config: dict[str, Any]) -> Path:
    return Path(config["paths"]["model_dir"]) / "ks_ae_metrics.json"


def _global_metrics_path(config: dict[str, Any]) -> Path:
    return Path(config["paths"]["model_dir"]) / "ks_global_ae_metrics.json"


def _diagnostics_path(config: dict[str, Any]) -> Path:
    return Path(config["paths"]["model_dir"]) / "ks_ae_diagnostics.npz"


def _global_diagnostics_path(config: dict[str, Any]) -> Path:
    return Path(config["paths"]["model_dir"]) / "ks_global_ae_diagnostics.npz"


def _spectra_path(config: dict[str, Any]) -> Path:
    return Path(config["paths"]["model_dir"]) / "ks_ae_spectra_summary.npz"


def _global_spectra_path(config: dict[str, Any]) -> Path:
    return Path(config["paths"]["model_dir"]) / "ks_global_ae_spectra_summary.npz"


def load_autoencoder(path: str | Path, device: str = "cpu") -> tuple[KSConvAutoencoder, float, float]:
    payload = torch.load(Path(path), map_location=device, weights_only=False)
    model = KSConvAutoencoder(
        n=int(payload["n"]),
        latent_channels=int(payload["latent_channels"]),
        kz=int(payload["kz"]),
        base_channels=int(payload["base_channels"]),
    )
    model.load_state_dict(payload["state_dict"])
    model.to(device)
    model.eval()
    return model, float(payload["mean"]), float(payload["std"])


def load_global_autoencoder(path: str | Path, device: str = "cpu") -> tuple[nn.Module, float, float]:
    payload = torch.load(Path(path), map_location=device, weights_only=False)
    architecture = str(payload.get("architecture", "conv_dense"))
    if architecture == "pca_global":
        model: nn.Module = KSGlobalPCAAutoencoder(
            n=int(payload["n"]),
            latent_dim=int(payload["latent_dim"]),
        )
    else:
        model = KSGlobalAutoencoder(
            n=int(payload["n"]),
            latent_dim=int(payload["latent_dim"]),
            bottleneck_grid=int(payload["bottleneck_grid"]),
            base_channels=int(payload["base_channels"]),
        )
    model.load_state_dict(payload["state_dict"])
    model.to(device)
    model.eval()
    return model, float(payload["mean"]), float(payload["std"])


def encode_physical(
    model: KSConvAutoencoder,
    states: np.ndarray,
    mean: float,
    std: float,
    device: str,
) -> np.ndarray:
    x = torch.as_tensor(((states.astype(np.float32) - mean) / std), device=device).unsqueeze(1)
    model.eval()
    with torch.no_grad():
        z = model.encode(x).detach().cpu().numpy()
    return z.reshape(states.shape[0], -1).astype(np.float64)


def encode_global_physical(
    model: nn.Module,
    states: np.ndarray,
    mean: float,
    std: float,
    device: str,
) -> np.ndarray:
    x = torch.as_tensor(((states.astype(np.float32) - mean) / std), device=device).unsqueeze(1)
    model.eval()
    with torch.no_grad():
        z = model.encode(x).detach().cpu().numpy()
    return z.astype(np.float64)


def decode_latent(
    model: KSConvAutoencoder,
    latent_flat: np.ndarray,
    mean: float,
    std: float,
    device: str,
) -> np.ndarray:
    n_members = int(latent_flat.shape[0])
    z = torch.as_tensor(
        latent_flat.reshape(n_members, model.latent_channels, model.kz).astype(np.float32),
        device=device,
    )
    model.eval()
    with torch.no_grad():
        decoded = model.decode(z).squeeze(1).detach().cpu().numpy()
    return (decoded * std + mean).astype(np.float64)


def decode_global_latent(
    model: nn.Module,
    latent: np.ndarray,
    mean: float,
    std: float,
    device: str,
) -> np.ndarray:
    z = torch.as_tensor(latent.astype(np.float32), device=device)
    model.eval()
    with torch.no_grad():
        decoded = model.decode(z).squeeze(1).detach().cpu().numpy()
    return (decoded * std + mean).astype(np.float64)


def _spectral_loss(recon: torch.Tensor, truth: torch.Tensor, max_mode: int) -> torch.Tensor:
    recon_hat = torch.fft.rfft(recon.squeeze(1), dim=-1)
    truth_hat = torch.fft.rfft(truth.squeeze(1), dim=-1)
    max_idx = min(int(max_mode), recon_hat.shape[-1] - 1)
    if max_idx <= 1:
        return torch.zeros((), dtype=recon.dtype, device=recon.device)
    recon_log = torch.log1p(torch.abs(recon_hat[..., 1 : max_idx + 1]))
    truth_log = torch.log1p(torch.abs(truth_hat[..., 1 : max_idx + 1]))
    return nn.functional.mse_loss(recon_log, truth_log)


def _loss_terms(model: nn.Module, x: torch.Tensor, ae_cfg: dict[str, Any]) -> dict[str, torch.Tensor]:
    recon = model(x)
    mse = nn.functional.mse_loss(recon, x)
    derivative = nn.functional.mse_loss(recon - torch.roll(recon, shifts=1, dims=-1), x - torch.roll(x, shifts=1, dims=-1))
    spectrum = _spectral_loss(recon, x, int(ae_cfg.get("spectrum_loss_max_mode", x.shape[-1] // 3)))
    total = (
        mse
        + float(ae_cfg.get("derivative_loss_weight", 0.0)) * derivative
        + float(ae_cfg.get("spectrum_loss_weight", 0.0)) * spectrum
    )
    return {"total": total, "mse": mse, "derivative": derivative, "spectrum": spectrum}


def _evaluate_losses(model: nn.Module, loader: DataLoader, device: str, ae_cfg: dict[str, Any]) -> dict[str, float]:
    model.eval()
    totals = {"total": 0.0, "mse": 0.0, "derivative": 0.0, "spectrum": 0.0}
    count = 0
    with torch.no_grad():
        for x in loader:
            x = x.to(device)
            terms = _loss_terms(model, x, ae_cfg)
            batch = int(x.shape[0])
            for key in totals:
                totals[key] += float(terms[key].item()) * batch
            count += batch
    return {key: value / max(count, 1) for key, value in totals.items()}


def _save_checkpoint(
    model: KSConvAutoencoder,
    config: dict[str, Any],
    mean: float,
    std: float,
    best_val_loss: float,
    phase: str,
    epoch: int,
) -> None:
    ae = config["ae"]
    torch.save(
        {
            "state_dict": model.state_dict(),
            "n": int(config["model"]["N"]),
            "length": float(config["model"]["length"]),
            "latent_channels": int(ae["latent_channels"]),
            "kz": int(ae["Kz"]),
            "base_channels": int(ae["base_channels"]),
            "mean": float(mean),
            "std": float(std),
            "best_val_loss": float(best_val_loss),
            "best_epoch": int(epoch),
            "train_phase": phase,
        },
        checkpoint_path(config),
    )


def _save_global_checkpoint(
    model: nn.Module,
    config: dict[str, Any],
    mean: float,
    std: float,
    best_val_loss: float,
    phase: str,
    epoch: int,
) -> None:
    ae = config["global_ae"]
    torch.save(
        {
            "state_dict": model.state_dict(),
            "n": int(config["model"]["N"]),
            "length": float(config["model"]["length"]),
            "latent_dim": int(ae["latent_dim"]),
            "architecture": str(ae.get("architecture", "conv_dense")),
            "bottleneck_grid": int(ae.get("bottleneck_grid", 0)),
            "base_channels": int(ae.get("base_channels", 0)),
            "mean": float(mean),
            "std": float(std),
            "best_val_loss": float(best_val_loss),
            "best_epoch": int(epoch),
            "train_phase": phase,
        },
        global_checkpoint_path(config),
    )


def _train_phase(
    model: KSConvAutoencoder,
    train_loader: DataLoader,
    val_loader: DataLoader,
    device: str,
    config: dict[str, Any],
    mean: float,
    std: float,
    phase: str,
    learning_rate: float,
    epochs: int,
    patience: int,
    initial_best: float,
) -> dict[str, Any]:
    ae_cfg = config["ae"]
    optimizer = torch.optim.AdamW(
        model.parameters(),
        lr=float(learning_rate),
        weight_decay=float(ae_cfg["weight_decay"]),
    )
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
        optimizer,
        mode="min",
        factor=float(ae_cfg.get("scheduler_factor", 0.5)),
        patience=int(ae_cfg.get("scheduler_patience", 5)),
        min_lr=float(ae_cfg.get("min_learning_rate", 1.0e-5)),
    )
    best_val = float(initial_best)
    best_epoch = 0
    epochs_since_improve = 0
    history: list[dict[str, Any]] = []
    start = time.time()
    grad_clip = float(ae_cfg.get("gradient_clip_norm", 0.0))
    for epoch in range(1, int(epochs) + 1):
        model.train()
        train_totals = {"total": 0.0, "mse": 0.0, "derivative": 0.0, "spectrum": 0.0}
        train_count = 0
        max_grad_norm = 0.0
        for x in train_loader:
            x = x.to(device)
            optimizer.zero_grad(set_to_none=True)
            terms = _loss_terms(model, x, ae_cfg)
            if not torch.isfinite(terms["total"]):
                raise FloatingPointError(f"Non-finite KS AE loss in phase {phase}, epoch {epoch}")
            terms["total"].backward()
            if grad_clip > 0.0:
                grad_norm = float(torch.nn.utils.clip_grad_norm_(model.parameters(), grad_clip).item())
                max_grad_norm = max(max_grad_norm, grad_norm)
            optimizer.step()
            batch = int(x.shape[0])
            for key in train_totals:
                train_totals[key] += float(terms[key].item()) * batch
            train_count += batch
        train_metrics = {key: value / max(train_count, 1) for key, value in train_totals.items()}
        val_metrics = _evaluate_losses(model, val_loader, device, ae_cfg)
        scheduler.step(val_metrics["total"])
        row = {
            "phase": phase,
            "epoch": int(epoch),
            "learning_rate": float(optimizer.param_groups[0]["lr"]),
            "train_loss": float(train_metrics["total"]),
            "train_mse": float(train_metrics["mse"]),
            "train_derivative_loss": float(train_metrics["derivative"]),
            "train_spectrum_loss": float(train_metrics["spectrum"]),
            "val_loss": float(val_metrics["total"]),
            "val_mse": float(val_metrics["mse"]),
            "val_normalized_rmse": float(np.sqrt(val_metrics["mse"])),
            "val_derivative_loss": float(val_metrics["derivative"]),
            "val_spectrum_loss": float(val_metrics["spectrum"]),
            "max_grad_norm": float(max_grad_norm),
        }
        history.append(row)
        if val_metrics["total"] < best_val:
            best_val = float(val_metrics["total"])
            best_epoch = int(epoch)
            epochs_since_improve = 0
            _save_checkpoint(model, config, mean, std, best_val, phase, epoch)
        else:
            epochs_since_improve += 1
        if epochs_since_improve >= int(patience):
            break
    return {
        "phase": phase,
        "best_val_loss": float(best_val),
        "best_epoch": int(best_epoch),
        "epochs_run": len(history),
        "elapsed": format_seconds(time.time() - start),
        "history": history,
    }


def _train_global_phase(
    model: KSGlobalAutoencoder,
    train_loader: DataLoader,
    val_loader: DataLoader,
    device: str,
    config: dict[str, Any],
    mean: float,
    std: float,
    phase: str,
    learning_rate: float,
    epochs: int,
    patience: int,
    initial_best: float,
) -> dict[str, Any]:
    ae_cfg = config["global_ae"]
    optimizer = torch.optim.AdamW(
        model.parameters(),
        lr=float(learning_rate),
        weight_decay=float(ae_cfg["weight_decay"]),
    )
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
        optimizer,
        mode="min",
        factor=float(ae_cfg.get("scheduler_factor", 0.5)),
        patience=int(ae_cfg.get("scheduler_patience", 5)),
        min_lr=float(ae_cfg.get("min_learning_rate", 1.0e-5)),
    )
    best_val = float(initial_best)
    best_epoch = 0
    epochs_since_improve = 0
    history: list[dict[str, Any]] = []
    start = time.time()
    grad_clip = float(ae_cfg.get("gradient_clip_norm", 0.0))
    for epoch in range(1, int(epochs) + 1):
        model.train()
        train_totals = {"total": 0.0, "mse": 0.0, "derivative": 0.0, "spectrum": 0.0}
        train_count = 0
        max_grad_norm = 0.0
        for x in train_loader:
            x = x.to(device)
            optimizer.zero_grad(set_to_none=True)
            terms = _loss_terms(model, x, ae_cfg)
            if not torch.isfinite(terms["total"]):
                raise FloatingPointError(f"Non-finite KS global AE loss in phase {phase}, epoch {epoch}")
            terms["total"].backward()
            if grad_clip > 0.0:
                grad_norm = float(torch.nn.utils.clip_grad_norm_(model.parameters(), grad_clip).item())
                max_grad_norm = max(max_grad_norm, grad_norm)
            optimizer.step()
            batch = int(x.shape[0])
            for key in train_totals:
                train_totals[key] += float(terms[key].item()) * batch
            train_count += batch
        train_metrics = {key: value / max(train_count, 1) for key, value in train_totals.items()}
        val_metrics = _evaluate_losses(model, val_loader, device, ae_cfg)
        scheduler.step(val_metrics["total"])
        row = {
            "phase": phase,
            "epoch": int(epoch),
            "learning_rate": float(optimizer.param_groups[0]["lr"]),
            "train_loss": float(train_metrics["total"]),
            "train_mse": float(train_metrics["mse"]),
            "train_derivative_loss": float(train_metrics["derivative"]),
            "train_spectrum_loss": float(train_metrics["spectrum"]),
            "val_loss": float(val_metrics["total"]),
            "val_mse": float(val_metrics["mse"]),
            "val_normalized_rmse": float(np.sqrt(val_metrics["mse"])),
            "val_derivative_loss": float(val_metrics["derivative"]),
            "val_spectrum_loss": float(val_metrics["spectrum"]),
            "max_grad_norm": float(max_grad_norm),
        }
        history.append(row)
        if val_metrics["total"] < best_val:
            best_val = float(val_metrics["total"])
            best_epoch = int(epoch)
            epochs_since_improve = 0
            _save_global_checkpoint(model, config, mean, std, best_val, phase, epoch)
        else:
            epochs_since_improve += 1
        if epochs_since_improve >= int(patience):
            break
    return {
        "phase": phase,
        "best_val_loss": float(best_val),
        "best_epoch": int(best_epoch),
        "epochs_run": len(history),
        "elapsed": format_seconds(time.time() - start),
        "history": history,
    }


def _reconstruct(model: nn.Module, states: np.ndarray, mean: float, std: float, device: str, batch_size: int) -> np.ndarray:
    model.eval()
    outputs = []
    with torch.no_grad():
        for start in range(0, states.shape[0], int(batch_size)):
            batch = states[start : start + int(batch_size)].astype(np.float32)
            x = torch.as_tensor((batch - mean) / std, device=device).unsqueeze(1)
            recon = model(x).squeeze(1).detach().cpu().numpy()
            outputs.append(recon * std + mean)
    return np.concatenate(outputs, axis=0).astype(np.float32)


def _metrics_for_split(
    config: dict[str, Any],
    split_name: str,
    truth: np.ndarray,
    recon: np.ndarray,
    normalization_std: float,
) -> dict[str, float]:
    params = params_from_config(config)
    gates = config["gates"]
    truth_spec = power_spectrum(truth, params)["power_spectrum"]
    recon_spec = power_spectrum(recon, params)["power_spectrum"]
    low_k_max = int(gates["low_k_max"])
    active_k_max = min(int(gates["active_k_max"]), params.n // 3)
    normalized = (recon - truth) / float(normalization_std)
    return {
        f"{split_name}_normalized_reconstruction_rmse": float(np.sqrt(np.mean(normalized**2))),
        f"{split_name}_physical_rmse": float(np.sqrt(np.mean((recon - truth) ** 2))),
        f"{split_name}_mean_abs_error": float(abs(np.mean(recon) - np.mean(truth))),
        f"{split_name}_variance_rel_error": _relative_scalar_error(float(np.var(recon)), float(np.var(truth))),
        f"{split_name}_low_k_spectrum_rel_l1": _spectrum_rel_l1(recon_spec, truth_spec, 1, low_k_max),
        f"{split_name}_active_band_spectrum_rel_l1": _spectrum_rel_l1(recon_spec, truth_spec, 1, active_k_max),
    }


def _relative_scalar_error(value: float, reference: float) -> float:
    return float(abs(value - reference) / max(abs(reference), 1.0e-12))


def _quality_gates(config: dict[str, Any], diagnostics: dict[str, float]) -> dict[str, Any]:
    gates = config["gates"]
    split = str(gates.get("selection_split", "val"))
    if split not in {"val", "holdout"}:
        raise ValueError("gates.selection_split must be 'val' or 'holdout'.")
    nrmse = float(diagnostics[f"{split}_normalized_reconstruction_rmse"])
    low = float(diagnostics[f"{split}_low_k_spectrum_rel_l1"])
    active = float(diagnostics[f"{split}_active_band_spectrum_rel_l1"])
    finite = bool(diagnostics["val_outputs_finite"] and diagnostics["holdout_outputs_finite"])
    primary = bool(
        finite
        and nrmse <= float(gates["primary_normalized_rmse_max"])
        and low <= float(gates["primary_low_k_spectrum_rel_error_max"])
        and active <= float(gates["primary_active_band_spectrum_rel_error_max"])
    )
    backup = bool(
        finite
        and nrmse <= float(gates["backup_normalized_rmse_max"])
        and low <= float(gates["backup_low_k_spectrum_rel_error_max"])
        and active <= float(gates["backup_active_band_spectrum_rel_error_max"])
    )
    return {
        "primary_pass": primary,
        "backup_pass": backup,
        "selection_split": split,
        "thresholds": gates,
    }


def _write_diagnostics(
    config: dict[str, Any],
    model: KSConvAutoencoder,
    data: dict[str, np.ndarray],
    mean: float,
    std: float,
    device: str,
) -> dict[str, Any]:
    ae_cfg = config["ae"]
    batch_size = int(ae_cfg["batch_size"])
    val = data["val"]
    holdout = data["holdout"]
    val_recon = _reconstruct(model, val, mean, std, device, batch_size)
    holdout_recon = _reconstruct(model, holdout, mean, std, device, batch_size)
    diagnostics: dict[str, Any] = {}
    diagnostics.update(_metrics_for_split(config, "val", val, val_recon, std))
    diagnostics.update(_metrics_for_split(config, "holdout", holdout, holdout_recon, std))
    diagnostics["val_outputs_finite"] = bool(np.all(np.isfinite(val_recon)))
    diagnostics["holdout_outputs_finite"] = bool(np.all(np.isfinite(holdout_recon)))

    params = params_from_config(config)
    val_truth_spec = power_spectrum(val, params)
    val_recon_spec = power_spectrum(val_recon, params)
    holdout_truth_spec = power_spectrum(holdout, params)
    holdout_recon_spec = power_spectrum(holdout_recon, params)

    sample_count = min(int(ae_cfg["diagnostic_samples"]), int(val.shape[0]))
    sample = val[:sample_count].astype(np.float32)
    sample_recon = val_recon[:sample_count].astype(np.float32)
    x_norm = torch.as_tensor((sample - mean) / std, device=device).unsqueeze(1)
    model.eval()
    responses = []
    with torch.no_grad():
        z0 = model.encode(x_norm[:1])
        base = model.decode(z0).squeeze(1)
        center = model.kz // 2
        for channel in range(model.latent_channels):
            zp = z0.clone()
            zp[:, channel, center] += float(ae_cfg["diagnostic_perturbation"])
            diff = (model.decode(zp).squeeze(1) - base).detach().cpu().numpy()[0] * std
            responses.append(diff)

    np.savez_compressed(
        _diagnostics_path(config),
        truth=sample,
        reconstruction=sample_recon,
        latent_cell_responses=np.asarray(responses, dtype=np.float32),
        latent_cell_index=np.asarray([center], dtype=np.int64),
        latent_cell_physical_coord=np.asarray([center * model.n / model.kz], dtype=np.float32),
    )
    np.savez_compressed(
        _spectra_path(config),
        wavenumber=val_truth_spec["wavenumber"],
        val_truth_power_spectrum=val_truth_spec["power_spectrum"],
        val_reconstruction_power_spectrum=val_recon_spec["power_spectrum"],
        holdout_truth_power_spectrum=holdout_truth_spec["power_spectrum"],
        holdout_reconstruction_power_spectrum=holdout_recon_spec["power_spectrum"],
    )
    diagnostics["quality_gates"] = _quality_gates(config, diagnostics)
    return diagnostics


def _write_global_diagnostics(
    config: dict[str, Any],
    model: KSGlobalAutoencoder,
    data: dict[str, np.ndarray],
    mean: float,
    std: float,
    device: str,
) -> dict[str, Any]:
    ae_cfg = config["global_ae"]
    batch_size = int(ae_cfg["batch_size"])
    val = data["val"]
    holdout = data["holdout"]
    val_recon = _reconstruct(model, val, mean, std, device, batch_size)
    holdout_recon = _reconstruct(model, holdout, mean, std, device, batch_size)
    diagnostics: dict[str, Any] = {}
    diagnostics.update(_metrics_for_split(config, "val", val, val_recon, std))
    diagnostics.update(_metrics_for_split(config, "holdout", holdout, holdout_recon, std))
    diagnostics["val_outputs_finite"] = bool(np.all(np.isfinite(val_recon)))
    diagnostics["holdout_outputs_finite"] = bool(np.all(np.isfinite(holdout_recon)))

    params = params_from_config(config)
    val_truth_spec = power_spectrum(val, params)
    val_recon_spec = power_spectrum(val_recon, params)
    holdout_truth_spec = power_spectrum(holdout, params)
    holdout_recon_spec = power_spectrum(holdout_recon, params)

    sample_count = min(int(ae_cfg["diagnostic_samples"]), int(val.shape[0]))
    sample = val[:sample_count].astype(np.float32)
    sample_recon = val_recon[:sample_count].astype(np.float32)
    x_norm = torch.as_tensor((sample - mean) / std, device=device).unsqueeze(1)
    model.eval()
    responses = []
    with torch.no_grad():
        z0 = model.encode(x_norm[:1])
        base = model.decode(z0).squeeze(1)
        indices = np.asarray(
            [
                0,
                model.latent_dim // 4,
                model.latent_dim // 2,
                (3 * model.latent_dim) // 4,
            ],
            dtype=np.int64,
        )
        for index in indices:
            zp = z0.clone()
            zp[:, int(index)] += float(ae_cfg["diagnostic_perturbation"])
            diff = (model.decode(zp).squeeze(1) - base).detach().cpu().numpy()[0] * std
            responses.append(diff)

    np.savez_compressed(
        _global_diagnostics_path(config),
        truth=sample,
        reconstruction=sample_recon,
        latent_unit_responses=np.asarray(responses, dtype=np.float32),
        latent_unit_index=indices,
    )
    np.savez_compressed(
        _global_spectra_path(config),
        wavenumber=val_truth_spec["wavenumber"],
        val_truth_power_spectrum=val_truth_spec["power_spectrum"],
        val_reconstruction_power_spectrum=val_recon_spec["power_spectrum"],
        holdout_truth_power_spectrum=holdout_truth_spec["power_spectrum"],
        holdout_reconstruction_power_spectrum=holdout_recon_spec["power_spectrum"],
    )
    diagnostics["quality_gates"] = _quality_gates(config, diagnostics)
    return diagnostics


def _parameter_count(model: nn.Module) -> int:
    params = int(sum(param.numel() for param in model.parameters()))
    if params > 0:
        return params
    return int(sum(buffer.numel() for buffer in model.buffers()))


def train_autoencoder(config: dict[str, Any], overwrite_data: bool = False, overwrite_ae: bool = False) -> Path:
    ensure_dir(config["paths"]["model_dir"])
    ensure_dir(config["paths"]["figures_dir"])
    out_path = checkpoint_path(config)
    metrics_path = _metrics_path(config)
    if out_path.exists() and metrics_path.exists() and not overwrite_ae:
        return out_path

    generate_autoencoder_data(config, overwrite=overwrite_data)
    data_path = Path(config["paths"]["data_dir"]) / "ks_ae_data.npz"
    payload = np.load(data_path)
    data = {
        "train": payload["train"],
        "val": payload["val"],
        "holdout": payload["holdout"],
    }
    mean = float(payload["mean"])
    std = float(payload["std"])
    ae_cfg = config["ae"]
    set_seed(int(config["seeds"].get("ae_train", 44001)))
    device = best_device(str(ae_cfg.get("device", "auto")))
    model = KSConvAutoencoder(
        n=int(config["model"]["N"]),
        latent_channels=int(ae_cfg["latent_channels"]),
        kz=int(ae_cfg["Kz"]),
        base_channels=int(ae_cfg["base_channels"]),
    ).to(device)
    train_loader = DataLoader(
        KSDataset(data["train"], mean, std, augment_shift=True),
        batch_size=int(ae_cfg["batch_size"]),
        shuffle=True,
        num_workers=0,
    )
    val_loader = DataLoader(
        KSDataset(data["val"], mean, std, augment_shift=False),
        batch_size=int(ae_cfg["batch_size"]),
        shuffle=False,
        num_workers=0,
    )

    overall_start = time.time()
    phases = [
        _train_phase(
            model,
            train_loader,
            val_loader,
            device,
            config,
            mean,
            std,
            "main",
            float(ae_cfg["learning_rate"]),
            int(ae_cfg["epochs"]),
            int(ae_cfg["patience"]),
            float("inf"),
        )
    ]
    model, mean, std = load_autoencoder(out_path, device=device)
    diagnostics = _write_diagnostics(config, model, data, mean, std, device)
    if (
        bool(ae_cfg.get("finetune_if_backup_passes", True))
        and not bool(diagnostics["quality_gates"]["primary_pass"])
        and bool(diagnostics["quality_gates"]["backup_pass"])
    ):
        phases.append(
            _train_phase(
                model,
                train_loader,
                val_loader,
                device,
                config,
                mean,
                std,
                "finetune",
                float(ae_cfg["finetune_learning_rate"]),
                int(ae_cfg["finetune_epochs"]),
                int(ae_cfg["finetune_patience"]),
                float(phases[-1]["best_val_loss"]),
            )
        )
        model, mean, std = load_autoencoder(out_path, device=device)
        diagnostics = _write_diagnostics(config, model, data, mean, std, device)

    history = [row for phase in phases for row in phase["history"]]
    metrics = {
        "checkpoint": out_path,
        "data_path": data_path,
        "device": device,
        "architecture": str(ae_cfg.get("architecture", "conv_dense")),
        "elapsed": format_seconds(time.time() - overall_start),
        "phases": phases,
        "history": history,
        "diagnostics": {k: v for k, v in diagnostics.items() if k != "quality_gates"},
        "quality_gates": diagnostics["quality_gates"],
        "mean": mean,
        "std": std,
        "latent_shape": [int(ae_cfg["latent_channels"]), int(ae_cfg["Kz"])],
        "latent_dimension": int(ae_cfg["latent_channels"]) * int(ae_cfg["Kz"]),
        "physical_dimension": int(config["model"]["N"]),
        "parameter_count": _parameter_count(model),
        "compression_ratio": float(int(config["model"]["N"]) / (int(ae_cfg["latent_channels"]) * int(ae_cfg["Kz"]))),
    }
    save_json(metrics, metrics_path)
    save_json(
        {
            "checkpoint": out_path,
            "quality_gates": diagnostics["quality_gates"],
            "diagnostics": metrics["diagnostics"],
        },
        Path(config["paths"]["figures_dir"]) / "ks_ae_quality_summary.json",
    )
    return out_path


def train_global_autoencoder(config: dict[str, Any], overwrite_data: bool = False, overwrite_ae: bool = False) -> Path:
    ensure_dir(config["paths"]["model_dir"])
    ensure_dir(config["paths"]["figures_dir"])
    out_path = global_checkpoint_path(config)
    metrics_path = _global_metrics_path(config)
    if out_path.exists() and metrics_path.exists() and not overwrite_ae:
        return out_path

    if "global_ae" not in config:
        raise KeyError("Config must define global_ae to train the global KS autoencoder.")
    ae_cfg = config["global_ae"]
    if str(ae_cfg.get("architecture", "conv_dense")) == "pca_global":
        return train_global_pca_autoencoder(config, overwrite_data=overwrite_data, overwrite_ae=overwrite_ae)
    generate_autoencoder_data(config, overwrite=overwrite_data)
    data_path = Path(config["paths"]["data_dir"]) / "ks_ae_data.npz"
    payload = np.load(data_path)
    data = {
        "train": payload["train"],
        "val": payload["val"],
        "holdout": payload["holdout"],
    }
    mean = float(payload["mean"])
    std = float(payload["std"])
    set_seed(int(config["seeds"].get("global_ae_train", config["seeds"].get("ae_train", 44001) + 17)))
    device = best_device(str(ae_cfg.get("device", config.get("ae", {}).get("device", "auto"))))
    model = KSGlobalAutoencoder(
        n=int(config["model"]["N"]),
        latent_dim=int(ae_cfg["latent_dim"]),
        bottleneck_grid=int(ae_cfg["bottleneck_grid"]),
        base_channels=int(ae_cfg["base_channels"]),
    ).to(device)
    train_loader = DataLoader(
        KSDataset(data["train"], mean, std, augment_shift=True),
        batch_size=int(ae_cfg["batch_size"]),
        shuffle=True,
        num_workers=0,
    )
    val_loader = DataLoader(
        KSDataset(data["val"], mean, std, augment_shift=False),
        batch_size=int(ae_cfg["batch_size"]),
        shuffle=False,
        num_workers=0,
    )

    overall_start = time.time()
    phases = [
        _train_global_phase(
            model,
            train_loader,
            val_loader,
            device,
            config,
            mean,
            std,
            "main",
            float(ae_cfg["learning_rate"]),
            int(ae_cfg["epochs"]),
            int(ae_cfg["patience"]),
            float("inf"),
        )
    ]
    model, mean, std = load_global_autoencoder(out_path, device=device)
    diagnostics = _write_global_diagnostics(config, model, data, mean, std, device)
    if (
        bool(ae_cfg.get("finetune_if_backup_passes", True))
        and not bool(diagnostics["quality_gates"]["primary_pass"])
        and bool(diagnostics["quality_gates"]["backup_pass"])
    ):
        phases.append(
            _train_global_phase(
                model,
                train_loader,
                val_loader,
                device,
                config,
                mean,
                std,
                "finetune",
                float(ae_cfg["finetune_learning_rate"]),
                int(ae_cfg["finetune_epochs"]),
                int(ae_cfg["finetune_patience"]),
                float(phases[-1]["best_val_loss"]),
            )
        )
        model, mean, std = load_global_autoencoder(out_path, device=device)
        diagnostics = _write_global_diagnostics(config, model, data, mean, std, device)

    history = [row for phase in phases for row in phase["history"]]
    metrics = {
        "checkpoint": out_path,
        "data_path": data_path,
        "device": device,
        "architecture": str(ae_cfg.get("architecture", "conv_dense")),
        "elapsed": format_seconds(time.time() - overall_start),
        "phases": phases,
        "history": history,
        "diagnostics": {k: v for k, v in diagnostics.items() if k != "quality_gates"},
        "quality_gates": diagnostics["quality_gates"],
        "mean": mean,
        "std": std,
        "latent_shape": [int(ae_cfg["latent_dim"])],
        "latent_dimension": int(ae_cfg["latent_dim"]),
        "bottleneck_grid": int(ae_cfg["bottleneck_grid"]),
        "base_channels": int(ae_cfg["base_channels"]),
        "physical_dimension": int(config["model"]["N"]),
        "parameter_count": _parameter_count(model),
        "compression_ratio": float(int(config["model"]["N"]) / int(ae_cfg["latent_dim"])),
    }
    save_json(metrics, metrics_path)
    save_json(
        {
            "checkpoint": out_path,
            "quality_gates": diagnostics["quality_gates"],
            "diagnostics": metrics["diagnostics"],
        },
        Path(config["paths"]["figures_dir"]) / "ks_global_ae_quality_summary.json",
    )
    return out_path


def train_global_pca_autoencoder(
    config: dict[str, Any],
    overwrite_data: bool = False,
    overwrite_ae: bool = False,
) -> Path:
    ensure_dir(config["paths"]["model_dir"])
    ensure_dir(config["paths"]["figures_dir"])
    out_path = global_checkpoint_path(config)
    metrics_path = _global_metrics_path(config)
    if out_path.exists() and metrics_path.exists() and not overwrite_ae:
        return out_path

    generate_autoencoder_data(config, overwrite=overwrite_data)
    data_path = Path(config["paths"]["data_dir"]) / "ks_ae_data.npz"
    payload = np.load(data_path)
    data = {
        "train": payload["train"],
        "val": payload["val"],
        "holdout": payload["holdout"],
    }
    mean = float(payload["mean"])
    std = float(payload["std"])
    ae_cfg = config["global_ae"]
    latent_dim = int(ae_cfg["latent_dim"])
    device = best_device(str(ae_cfg.get("device", config.get("ae", {}).get("device", "auto"))))

    start = time.time()
    train_norm = ((data["train"].astype(np.float64) - mean) / std).astype(np.float64)
    pca_mean = np.mean(train_norm, axis=0)
    centered = train_norm - pca_mean[None, :]
    covariance = (centered.T @ centered) / max(int(centered.shape[0]) - 1, 1)
    eigvals, eigvecs = np.linalg.eigh(covariance)
    order = np.argsort(eigvals)[::-1]
    selected = order[:latent_dim]
    total_variance = float(np.sum(np.maximum(eigvals, 0.0)))
    explained = float(np.sum(np.maximum(eigvals[selected], 0.0)) / max(total_variance, 1.0e-20))

    model = KSGlobalPCAAutoencoder(n=int(config["model"]["N"]), latent_dim=latent_dim).to(device)
    with torch.no_grad():
        model.pca_mean.copy_(torch.as_tensor(pca_mean.astype(np.float32), device=device))
        model.components.copy_(torch.as_tensor(eigvecs[:, selected].T.astype(np.float32), device=device))

    _save_global_checkpoint(model, config, mean, std, 0.0, "pca", 1)
    model, mean, std = load_global_autoencoder(out_path, device=device)
    diagnostics = _write_global_diagnostics(config, model, data, mean, std, device)
    elapsed = format_seconds(time.time() - start)
    metrics = {
        "checkpoint": out_path,
        "data_path": data_path,
        "device": device,
        "elapsed": elapsed,
        "architecture": "pca_global",
        "explained_variance_ratio": explained,
        "phases": [
            {
                "phase": "pca",
                "best_val_loss": 0.0,
                "best_epoch": 1,
                "epochs_run": 1,
                "elapsed": elapsed,
                "explained_variance_ratio": explained,
                "history": [],
            }
        ],
        "history": [],
        "diagnostics": {k: v for k, v in diagnostics.items() if k != "quality_gates"},
        "quality_gates": diagnostics["quality_gates"],
        "mean": mean,
        "std": std,
        "latent_shape": [latent_dim],
        "latent_dimension": latent_dim,
        "physical_dimension": int(config["model"]["N"]),
        "parameter_count": _parameter_count(model),
        "compression_ratio": float(int(config["model"]["N"]) / latent_dim),
    }
    save_json(metrics, metrics_path)
    save_json(
        {
            "checkpoint": out_path,
            "architecture": "pca_global",
            "explained_variance_ratio": explained,
            "quality_gates": diagnostics["quality_gates"],
            "diagnostics": metrics["diagnostics"],
        },
        Path(config["paths"]["figures_dir"]) / "ks_global_ae_quality_summary.json",
    )
    return out_path


def compare_autoencoder_quality(config: dict[str, Any]) -> Path:
    spatial_path = _metrics_path(config)
    global_path = _global_metrics_path(config)
    if not spatial_path.exists():
        raise FileNotFoundError(f"Spatial AE metrics are missing: {spatial_path}")
    if not global_path.exists():
        raise FileNotFoundError(f"Global AE metrics are missing: {global_path}")
    import json

    with spatial_path.open("r", encoding="utf-8") as f:
        spatial = json.load(f)
    with global_path.open("r", encoding="utf-8") as f:
        global_ = json.load(f)
    gates = config.get("gates", {})
    split = str(gates.get("selection_split", "val"))
    key = f"{split}_normalized_reconstruction_rmse"
    spatial_nrmse = float(spatial["diagnostics"][key])
    global_nrmse = float(global_["diagnostics"][key])
    best = min(spatial_nrmse, global_nrmse)
    worst = max(spatial_nrmse, global_nrmse)
    rel_ratio = float(worst / max(best, 1.0e-12))
    abs_gap = float(worst - best)
    max_ratio = float(gates.get("comparability_max_nrmse_ratio", 1.25))
    max_abs_gap = float(gates.get("comparability_max_abs_nrmse_gap", 0.005))
    spatial_pass = bool(
        spatial.get("quality_gates", {}).get("primary_pass", False)
        or spatial.get("quality_gates", {}).get("backup_pass", False)
    )
    global_pass = bool(
        global_.get("quality_gates", {}).get("primary_pass", False)
        or global_.get("quality_gates", {}).get("backup_pass", False)
    )
    comparable = bool(spatial_pass and global_pass and (rel_ratio <= max_ratio or abs_gap <= max_abs_gap))
    summary = {
        "selection_split": split,
        "spatial_checkpoint": spatial.get("checkpoint"),
        "global_checkpoint": global_.get("checkpoint"),
        "spatial_nrmse": spatial_nrmse,
        "global_nrmse": global_nrmse,
        "nrmse_ratio_worst_over_best": rel_ratio,
        "nrmse_abs_gap": abs_gap,
        "spatial_gate_pass": spatial_pass,
        "global_gate_pass": global_pass,
        "comparable": comparable,
        "thresholds": {
            "max_ratio": max_ratio,
            "max_abs_gap": max_abs_gap,
        },
        "spatial_parameter_count": spatial.get("parameter_count"),
        "global_parameter_count": global_.get("parameter_count"),
        "spatial_elapsed": spatial.get("elapsed"),
        "global_elapsed": global_.get("elapsed"),
    }
    out_path = Path(config["paths"]["figures_dir"]) / "ks_ae_comparison_summary.json"
    save_json(summary, out_path)
    if not comparable:
        raise RuntimeError(f"Spatial/global AE quality is not comparable; see {out_path}")
    return out_path



from __future__ import annotations

import csv
import json
import math
import os
import shutil
from pathlib import Path

import numpy as np


ROOT = Path(r"D:\LDA")
PIC = ROOT / "PIC"
ORGANIZED = PIC / "Organized"

LATENT_TUNING_CSV = (
    ROOT
    / "outputs"
    / "PATH_NAME"
    / "tuning"
    / "method_specific_tuning_results.csv"
)
PHY_TUNING_CSV = (
    ROOT
    / "outputs"
    / "PATH_NAME"
    / "figures"
    / "test_oracle_phy_tuning_results.csv"
)
NORM_SOURCE_JSON = (
    ROOT
    / "outputs"
    / "PATH_NAME"
    / "results"
    / "N20_current_ae_loc_method_tuning_tuning"
    / "test_SLA_256_inf1p10_rad192p00_summary.json"
)

FIG_PATH = PIC / "Fig10_N20_method_specific_tuning_heatmaps.png"
FIG_PDF = PIC / "Fig10_N20_method_specific_tuning_heatmaps.pdf"
SUMMARY_CSV = PIC / "Fig10_N20_method_specific_tuning_heatmaps_values.csv"

METHODS = [
    ("GLA_256", "GLA_256 - PHY"),
    ("SLA_256", "SLA_256 - PHY"),
    ("SLA_128", "SLA_128 - PHY"),
]


def read_rows(path: Path) -> list[dict[str, str]]:
    with path.open("r", newline="", encoding="utf-8") as f:
        return list(csv.DictReader(f))


def as_float(text: str) -> float:
    try:
        return float(text)
    except (TypeError, ValueError):
        return math.nan


def rmse_std() -> float:
    with NORM_SOURCE_JSON.open("r", encoding="utf-8") as f:
        data = json.load(f)
    return float(data["normalization_std"])


def build_surface(
    rows: list[dict[str, str]],
    method_label: str | None,
    inflations: list[float],
    radii: list[float],
) -> tuple[np.ndarray, np.ndarray]:
    values = np.full((len(inflations), len(radii)), np.nan, dtype=float)
    unstable = np.zeros((len(inflations), len(radii)), dtype=bool)
    infl_index = {round(v, 8): i for i, v in enumerate(inflations)}
    rad_index = {round(v, 8): i for i, v in enumerate(radii)}

    for row in rows:
        if method_label is not None and row.get("method_label") != method_label:
            continue
        inf = as_float(row.get("inflation", ""))
        rad = as_float(row.get("loc_radius", ""))
        if not math.isfinite(inf) or not math.isfinite(rad):
            continue
        i = infl_index.get(round(inf, 8))
        j = rad_index.get(round(rad, 8))
        if i is None or j is None:
            continue
        values[i, j] = as_float(row.get("mean_analysis_nrmse", ""))
        unstable[i, j] = row.get("stable", "").strip().lower() != "true"
    return values, unstable


def main() -> None:
    mpl_config = ROOT / ".mplconfig_fig34_tuning_heatmaps"
    mpl_config.mkdir(exist_ok=True)
    os.environ["MPLCONFIGDIR"] = str(mpl_config)

    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    from matplotlib.colors import TwoSlopeNorm

    latent_rows = read_rows(LATENT_TUNING_CSV)
    phy_rows = read_rows(PHY_TUNING_CSV)

    inflations = sorted({as_float(r["inflation"]) for r in latent_rows if r.get("inflation")})
    radii = sorted({as_float(r["loc_radius"]) for r in latent_rows if r.get("loc_radius")})
    norm_std = rmse_std()

    phy_nrmse, phy_unstable = build_surface(phy_rows, None, inflations, radii)
    if np.any(phy_unstable):
        raise RuntimeError("The PHY tuning surface contains unstable rows; this plot expects a stable PHY reference.")

    method_data: dict[str, dict[str, np.ndarray]] = {}
    finite_abs_max = 0.0
    summary_rows: list[dict[str, object]] = []
    for method_label, _title in METHODS:
        method_nrmse, method_unstable = build_surface(latent_rows, method_label, inflations, radii)
        diff_rmse = (method_nrmse - phy_nrmse) * norm_std
        diff_rmse[method_unstable | ~np.isfinite(method_nrmse)] = np.nan
        finite = diff_rmse[np.isfinite(diff_rmse)]
        if finite.size:
            finite_abs_max = max(finite_abs_max, float(np.max(np.abs(finite))))
        method_data[method_label] = {
            "diff_rmse": diff_rmse,
            "unstable": method_unstable | ~np.isfinite(method_nrmse),
            "method_nrmse": method_nrmse,
            "method_rmse": method_nrmse * norm_std,
        }

        for i, inf in enumerate(inflations):
            for j, rad in enumerate(radii):
                summary_rows.append(
                    {
                        "method_label": method_label,
                        "inflation": inf,
                        "loc_radius": rad,
                        "latent_analysis_nrmse": method_nrmse[i, j],
                        "phy_analysis_nrmse": phy_nrmse[i, j],
                        "analysis_rmse_difference_latent_minus_phy": diff_rmse[i, j],
                        "unstable_or_missing": bool(method_data[method_label]["unstable"][i, j]),
                    }
                )

    with SUMMARY_CSV.open("w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(
            f,
            fieldnames=[
                "method_label",
                "inflation",
                "loc_radius",
                "latent_analysis_nrmse",
                "phy_analysis_nrmse",
                "analysis_rmse_difference_latent_minus_phy",
                "unstable_or_missing",
            ],
        )
        writer.writeheader()
        writer.writerows(summary_rows)

    matplotlib.rcParams.update(
        {
            "font.family": "serif",
            "font.size": 11.0,
            "axes.linewidth": 1.0,
            "xtick.direction": "out",
            "ytick.direction": "out",
        }
    )

    color_limit = max(0.05, math.ceil(finite_abs_max * 100.0) / 100.0)
    norm = TwoSlopeNorm(vmin=-color_limit, vcenter=0.0, vmax=color_limit)
    cmap = plt.get_cmap("RdBu_r").copy()
    cmap.set_bad(color="0.72")

    fig, axes = plt.subplots(1, 3, figsize=(12.8, 3.9), constrained_layout=False)
    fig.subplots_adjust(left=0.075, right=0.895, bottom=0.18, top=0.88, wspace=0.10)

    im = None
    letters = ["(a)", "(b)", "(c)"]
    x = np.arange(len(radii) + 1)
    y = np.arange(len(inflations) + 1)
    for ax, (method_label, title), letter in zip(axes, METHODS, letters):
        diff = method_data[method_label]["diff_rmse"]
        unstable = method_data[method_label]["unstable"]
        im = ax.pcolormesh(x, y, diff, cmap=cmap, norm=norm, shading="flat", edgecolors="white", linewidth=0.55)
        ax.set_xticks(np.arange(len(radii)) + 0.5)
        ax.set_xticklabels([str(int(r)) for r in radii], rotation=35, ha="right")
        ax.set_yticks(np.arange(len(inflations)) + 0.5)
        ax.set_yticklabels([f"{v:.2f}".rstrip("0").rstrip(".") for v in inflations])
        ax.set_xlabel("Localization radius")
        ax.set_title(f"{letter} {title}", fontsize=13.0)
        if ax is axes[0]:
            ax.set_ylabel("Inflation")
        else:
            ax.set_ylabel("")
            ax.tick_params(axis="y", labelleft=False)

        ys, xs = np.where(unstable)
        for yy, xx in zip(ys, xs):
            ax.text(
                xx + 0.5,
                yy + 0.5,
                "×",
                ha="center",
                va="center",
                color="black",
                fontsize=11.0,
                fontweight="bold",
            )

        method_rmse = method_data[method_label]["method_rmse"]
        stable_rmse = np.where(unstable, np.nan, method_rmse)
        best_flat = int(np.nanargmin(stable_rmse)) if np.any(np.isfinite(stable_rmse)) else -1
        best_yy, best_xx = np.unravel_index(best_flat, stable_rmse.shape) if best_flat >= 0 else (-1, -1)
        for yy in range(len(inflations)):
            for xx in range(len(radii)):
                if unstable[yy, xx] or not np.isfinite(method_rmse[yy, xx]):
                    continue
                text = f"{method_rmse[yy, xx]:.2f}"
                if yy == best_yy and xx == best_xx:
                    text += "*"
                ax.text(
                    xx + 0.5,
                    yy + 0.5,
                    text,
                    ha="center",
                    va="center",
                    fontsize=7.4,
                    color="black",
                    fontweight="bold" if yy == best_yy and xx == best_xx else "normal",
                )

    if im is not None:
        cax = fig.add_axes([0.915, 0.23, 0.018, 0.62])
        cbar = fig.colorbar(im, cax=cax, extend="both")
        cbar.set_label("Analysis RMSE difference\n(latent - PHY)")

    fig.savefig(FIG_PATH, dpi=300)
    fig.savefig(FIG_PDF)
    plt.close(fig)

    if ORGANIZED.exists():
        shutil.copy2(FIG_PATH, ORGANIZED / FIG_PATH.name)
        shutil.copy2(FIG_PDF, ORGANIZED / FIG_PDF.name)
        shutil.copy2(FIG_PATH, ORGANIZED / "Fig10_N20_method_specific_tuning_heatmaps.png")
        shutil.copy2(FIG_PDF, ORGANIZED / "Fig10_N20_method_specific_tuning_heatmaps.pdf")
        shutil.copy2(SUMMARY_CSV, ORGANIZED / "Fig10_N20_method_specific_tuning_heatmaps_values.csv")

    print(FIG_PATH)
    print(FIG_PDF)
    print(SUMMARY_CSV)


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import math
import os
import random
from pathlib import Path
from typing import Any

import numpy as np
import yaml


def ensure_dir(path: str | Path) -> Path:
    """Create a directory if needed and return it as a Path."""
    out = Path(path)
    out.mkdir(parents=True, exist_ok=True)
    return out


def set_seed(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed)
    try:
        import torch

        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)
    except Exception:
        pass


def load_yaml(path: str | Path) -> dict[str, Any]:
    with Path(path).open("r", encoding="utf-8") as f:
        return yaml.safe_load(f)


def save_json(obj: Any, path: str | Path) -> None:
    path = Path(path)
    ensure_dir(path.parent)
    with path.open("w", encoding="utf-8") as f:
        json.dump(to_jsonable(obj), f, indent=2, sort_keys=True)


def to_jsonable(obj: Any) -> Any:
    if isinstance(obj, dict):
        return {str(k): to_jsonable(v) for k, v in obj.items()}
    if isinstance(obj, (list, tuple)):
        return [to_jsonable(v) for v in obj]
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, np.generic):
        return obj.item()
    if isinstance(obj, Path):
        return str(obj)
    return obj


def project_root() -> Path:
    return Path(__file__).resolve().parents[1]


def resolve_path(path: str | Path, base: str | Path | None = None) -> Path:
    p = Path(path)
    if p.is_absolute():
        return p
    return (Path(base) if base is not None else project_root()) / p


def configure_matplotlib_cache(output_dir: str | Path) -> None:
    cache_dir = ensure_dir(Path(output_dir) / "mplconfig")
    os.environ.setdefault("MPLCONFIGDIR", str(cache_dir))


def rmse(mean_state: np.ndarray, truth: np.ndarray) -> float:
    return float(np.sqrt(np.mean((mean_state - truth) ** 2)))


def ensemble_spread(ensemble: np.ndarray) -> float:
    if ensemble.shape[0] < 2:
        return float("nan")
    variance = np.var(ensemble, axis=0, ddof=1)
    return float(np.sqrt(np.mean(variance)))


def rescale_to_rmse(error: np.ndarray, target_rmse: float) -> np.ndarray:
    current = np.sqrt(np.mean(error**2))
    if current == 0:
        raise ValueError("Cannot rescale a zero error vector.")
    return error * (target_rmse / current)


def cyclic_distances(points: np.ndarray, ref: float, domain_size: int) -> np.ndarray:
    distances = np.abs(points.astype(float) - float(ref))
    return np.minimum(distances, domain_size - distances)


def gaspari_cohn(distances: np.ndarray, radius: float) -> np.ndarray:
    """Gaspari-Cohn taper with compact support at ``radius``.

    The usual Gaspari-Cohn polynomial has support at 2c. Here the public
    parameter is the zero-cutoff radius, so c = radius / 2.
    """
    distances = np.asarray(distances, dtype=float)
    if radius <= 0:
        return np.zeros_like(distances)

    r = distances / (radius / 2.0)
    out = np.zeros_like(r)

    mask1 = r <= 1.0
    x = r[mask1]
    out[mask1] = (
        1.0
        - (5.0 / 3.0) * x**2
        + (5.0 / 8.0) * x**3
        + 0.5 * x**4
        - 0.25 * x**5
    )

    mask2 = (r > 1.0) & (r < 2.0)
    x = r[mask2]
    out[mask2] = (
        4.0
        - 5.0 * x
        + (5.0 / 3.0) * x**2
        + (5.0 / 8.0) * x**3
        - 0.5 * x**4
        + (1.0 / 12.0) * x**5
        - (2.0 / (3.0 * x))
    )
    return np.clip(out, 0.0, 1.0)


def summarize_cycles(
    prior_rmse: np.ndarray,
    analysis_rmse: np.ndarray,
    prior_spread: np.ndarray,
    analysis_spread: np.ndarray,
    discard_cycles: int,
) -> dict[str, float]:
    sl = slice(discard_cycles, None)
    return {
        "discard_cycles": int(discard_cycles),
        "mean_prior_rmse": float(np.mean(prior_rmse[sl])),
        "mean_analysis_rmse": float(np.mean(analysis_rmse[sl])),
        "mean_prior_spread": float(np.mean(prior_spread[sl])),
        "mean_analysis_spread": float(np.mean(analysis_spread[sl])),
        "prior_spread_rmse_ratio": float(np.mean(prior_spread[sl]) / np.mean(prior_rmse[sl])),
        "analysis_spread_rmse_ratio": float(
            np.mean(analysis_spread[sl]) / np.mean(analysis_rmse[sl])
        ),
        "stable": bool(np.all(np.isfinite(analysis_rmse))),
    }


def exact_grid_std_anomalies(rng: np.random.Generator, n_members: int, k: int, std: float) -> np.ndarray:
    anomalies = rng.normal(0.0, 1.0, size=(n_members, k))
    anomalies -= anomalies.mean(axis=0, keepdims=True)
    sample_std = anomalies.std(axis=0, ddof=1, keepdims=True)
    sample_std = np.where(sample_std == 0.0, 1.0, sample_std)
    return anomalies / sample_std * std


def best_device(requested: str = "auto") -> str:
    if requested != "auto":
        return requested
    try:
        import torch

        return "cuda" if torch.cuda.is_available() else "cpu"
    except Exception:
        return "cpu"


def format_seconds(seconds: float) -> str:
    if seconds < 60:
        return f"{seconds:.1f}s"
    minutes = math.floor(seconds / 60)
    return f"{minutes}m {seconds - 60 * minutes:.1f}s"



from __future__ import annotations

import argparse
from pathlib import Path

import numpy as np

from example_common import (
    diagnostics,
    forecast_ensemble,
    global_latent_localized_ensrf,
    inflate_ensemble,
    load_checkpoint,
    make_initial_ensemble,
    make_params,
    make_truth,
)
from lda.ks import integrate_steps
from lda.ks_ae import KSGlobalAutoencoder
from lda.utils import ensure_dir


def main() -> None:
    parser = argparse.ArgumentParser(description="Example global-latent EnSRF with decoded physical localization.")
    parser.add_argument("--n", type=int, default=512)
    parser.add_argument("--length", type=float, default=None)
    parser.add_argument("--dt", type=float, default=0.05)
    parser.add_argument("--cycle-steps", type=int, default=40)
    parser.add_argument("--cycles", type=int, default=10)
    parser.add_argument("--members", type=int, default=20)
    parser.add_argument("--obs-stride", type=int, default=32)
    parser.add_argument("--obs-error", type=float, default=0.5)
    parser.add_argument("--inflation", type=float, default=1.10)
    parser.add_argument("--loc-radius", type=float, default=192.0)
    parser.add_argument("--latent-dim", type=int, default=256)
    parser.add_argument("--bottleneck-grid", type=int, default=64)
    parser.add_argument("--base-channels", type=int, default=16)
    parser.add_argument("--checkpoint", type=str, default=None)
    parser.add_argument("--spinup-steps", type=int, default=2000)
    parser.add_argument("--seed", type=int, default=3)
    parser.add_argument("--output", type=Path, default=Path("example_outputs/gla_ensrf_loc.npz"))
    args = parser.parse_args()

    params = make_params(args.n, args.length)
    rng = np.random.default_rng(args.seed)
    truth = make_truth(params, rng, args.spinup_steps, args.dt)
    ensemble = make_initial_ensemble(truth, params, rng, args.members, perturbation_std=0.5)
    obs_locs = np.arange(0, args.n, args.obs_stride)

    model = KSGlobalAutoencoder(
        n=args.n,
        latent_dim=args.latent_dim,
        bottleneck_grid=args.bottleneck_grid,
        base_channels=args.base_channels,
    )
    model = load_checkpoint(model, args.checkpoint)
    norm_mean = float(ensemble.mean())
    norm_std = float(max(ensemble.std(), 1.0e-6))

    prior_rmse = []
    analysis_rmse = []
    for _ in range(args.cycles):
        truth = integrate_steps(truth, args.cycle_steps, args.dt, params)
        ensemble = forecast_ensemble(ensemble, args.cycle_steps, args.dt, params)
        prior = ensemble.copy()
        obs = truth[obs_locs] + rng.normal(0.0, args.obs_error, size=len(obs_locs))
        ensemble = inflate_ensemble(ensemble, args.inflation)
        ensemble = global_latent_localized_ensrf(
            ensemble,
            obs,
            obs_locs,
            args.obs_error,
            args.loc_radius,
            model,
            norm_mean,
            norm_std,
        )
        d = diagnostics(prior, ensemble, truth)
        prior_rmse.append(d[0])
        analysis_rmse.append(d[1])

    ensure_dir(args.output.parent)
    np.savez_compressed(
        args.output,
        prior_rmse=np.asarray(prior_rmse),
        analysis_rmse=np.asarray(analysis_rmse),
        obs_locs=obs_locs,
        inflation=args.inflation,
        loc_radius=args.loc_radius,
        latent_dim=args.latent_dim,
    )
    print(f"saved {args.output}")


if __name__ == "__main__":
    main()



from pathlib import Path
import os
import sys


SCRIPT_DIR = Path(__file__).resolve().parent
ROOT = SCRIPT_DIR.parent
if str(ROOT) not in sys.path:
    sys.path.insert(0, str(ROOT))
MPLCONFIGDIR = SCRIPT_DIR.parent / ".mplconfig_pic_full_hovmoller"
MPLCONFIGDIR.mkdir(exist_ok=True)
for lock_file in MPLCONFIGDIR.glob("*.matplotlib-lock"):
    try:
        lock_file.unlink()
    except OSError:
        pass
os.environ.setdefault("MPLCONFIGDIR", str(MPLCONFIGDIR))

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LinearSegmentedColormap

from lda.ks import KSParams, integrate_steps, random_initial_state, saved_trajectory


OUTDIR = SCRIPT_DIR
PNG_PATH = OUTDIR / "Fig01_full_truth_hovmoller_T1600.png"
PDF_PATH = OUTDIR / "Fig01_full_truth_hovmoller_T1600.pdf"
FORECAST_TRUTH_PATH = OUTDIR / "Fig01_forecast_truth_from_DA_start_T0_to_T600_dt005.npz"
RGB_CMAP_PATH = Path(r"amwg_blueyellowred.rgb")
USE_AMWG_CMAP = False


def load_rgb_colormap(rgb_path):
    colors = []
    with Path(rgb_path).open("r", encoding="utf-8") as f:
        for raw_line in f:
            line = raw_line.strip()
            if not line or line.startswith(";") or line.startswith("#") or line.lower().startswith("ncolors"):
                continue
            parts = line.split()
            if len(parts) < 3:
                continue
            colors.append([float(parts[0]) / 255.0, float(parts[1]) / 255.0, float(parts[2]) / 255.0])
    if not colors:
        raise ValueError(f"No RGB colors found in {rgb_path}")
    return LinearSegmentedColormap.from_list(Path(rgb_path).stem, colors, N=256)


def main():
    params = KSParams(
        n=2048,
        length=256.0,
        linear_growth=1.0,
        third_order_dispersion=0.1,
        fifth_order_dispersion=0.1,
        fourth_order_damping=1.0,
        dealias=True,
        max_abs_state=1.0e6,
    )
    dt = 0.05
    truth_spinup_steps = 10000
    ensemble_spinup_steps = 10000
    da_steps = 8000
    forecast_extension_steps = 4000
    da_plus_forecast_steps = da_steps + forecast_extension_steps
    total_steps = truth_spinup_steps + ensemble_spinup_steps + da_plus_forecast_steps
    total_time = total_steps * dt
    save_interval = 0.25

    rng = np.random.default_rng(76001)
    field0 = random_initial_state(
        params,
        rng,
        std=0.5,
        spectral_slope=4.0,
        max_initial_mode=32,
    )
    truth = saved_trajectory(
        field0,
        total_time=total_time,
        save_interval=save_interval,
        dt=dt,
        params=params,
        include_initial=True,
    ).astype(np.float64)

    da_start_state = integrate_steps(
        field0,
        truth_spinup_steps + ensemble_spinup_steps,
        dt,
        params,
    )
    forecast_truth = saved_trajectory(
        da_start_state,
        total_time=da_plus_forecast_steps * dt,
        save_interval=dt,
        dt=dt,
        params=params,
        include_initial=True,
    )
    np.savez_compressed(
        FORECAST_TRUTH_PATH,
        truth=forecast_truth.astype(np.float32),
        dt=np.asarray(dt, dtype=np.float64),
        da_steps=np.asarray(da_steps, dtype=np.int64),
        forecast_extension_steps=np.asarray(forecast_extension_steps, dtype=np.int64),
        truth_spinup_steps=np.asarray(truth_spinup_steps, dtype=np.int64),
        ensemble_spinup_steps=np.asarray(ensemble_spinup_steps, dtype=np.int64),
    )

    n_save, n_grid = truth.shape
    length = params.length
    t = np.arange(n_save, dtype=np.float64) * save_interval
    x = np.arange(n_grid, dtype=np.float64)

    q01, q99 = np.percentile(truth, [1.0, 99.0])
    vmax = max(abs(q01), abs(q99))

    plt.rcParams.update(
        {
            "font.family": "DejaVu Sans",
            "font.size": 15,
            "mathtext.fontset": "dejavusans",
            "axes.labelsize": 19,
            "xtick.labelsize": 16,
            "ytick.labelsize": 16,
        }
    )
    cmap = load_rgb_colormap(RGB_CMAP_PATH) if USE_AMWG_CMAP and RGB_CMAP_PATH.exists() else "RdBu_r"
    fig, ax = plt.subplots(figsize=(8.2, 9.2), facecolor="white")
    im = ax.imshow(
        truth,
        origin="lower",
        aspect="auto",
        interpolation="nearest",
        extent=[x[0], x[-1], t[0], t[-1]],
        cmap=cmap,
        vmin=-vmax,
        vmax=vmax,
    )

    t_truth_spinup = truth_spinup_steps * dt
    t_da_start = (truth_spinup_steps + ensemble_spinup_steps) * dt
    line1 = ax.axhline(t_truth_spinup, color="0.05", lw=3.0, ls="--", alpha=0.95)
    line2 = ax.axhline(t_da_start, color="0.05", lw=3.0, ls=":", alpha=0.95)

    ax.set_xlabel("Physical grid index")
    ax.set_ylabel("Time")
    ax.set_xlim(0, n_grid - 1)
    ax.set_ylim(0, total_time)
    ax.set_xticks([0, 512, 1024, 1536, 2047])

    ax.legend(
        [line1, line2],
        [r"$t=500$: truth spin-up ends", r"$t=1000$: DA period begins"],
        loc="upper left",
        frameon=True,
        facecolor="white",
        edgecolor="none",
        framealpha=0.82,
        fontsize=10.5,
        handlelength=2.7,
        borderpad=0.45,
        labelspacing=0.35,
    )

    cbar = fig.colorbar(im, ax=ax, pad=0.018)
    cbar.set_label(r"State $u(x,t)$", fontsize=17)
    cbar.ax.tick_params(labelsize=15)

    fig.tight_layout()
    fig.savefig(PNG_PATH, dpi=300, bbox_inches="tight")
    fig.savefig(PDF_PATH, bbox_inches="tight")
    print(f"saved {PNG_PATH}")
    print(f"saved {PDF_PATH}")
    print(f"saved {FORECAST_TRUTH_PATH}")


if __name__ == "__main__":
    main()



from __future__ import annotations

import csv
import os
from pathlib import Path

import numpy as np


ROOT = Path(r"D:\LDA")
PIC = ROOT / "PIC"
MPLCONFIGDIR = ROOT / ".mplconfig_pic_fig30f_early_late"
MPLCONFIGDIR.mkdir(exist_ok=True)
for lock_file in MPLCONFIGDIR.glob("*.matplotlib-lock"):
    try:
        lock_file.unlink()
    except OSError:
        pass
os.environ["MPLCONFIGDIR"] = str(MPLCONFIGDIR)

RESULT_DIR = (
    ROOT
    / "outputs"
    / "PATH_NAME"
    / "results"
    / "phyloc_inflation_shared"
)
TEST_CASE_PATH = (
    ROOT
    / "outputs"
    / "PATH_NAME"
    / "data"
    / "test_case.npz"
)

OUT_BASE = PIC / "Fig07_cycles_early_late_combined_cumulative"
OUT_SUMMARY = OUT_BASE.with_name(OUT_BASE.name + "_summary.csv")

DEALIAS_FRAC = 2.0 / 3.0
AXIS_LABEL_FONT_SIZE = 12.2
PANEL_TAG_FONT_SIZE = 13.0
SPECTRUM_Y_MIN = 1.0e-18
SPECTRUM_Y_SPLIT = 1.0e-7
SPECTRUM_Y_MAX = 1.0e-1
SPECTRUM_LOW_FRACTION = 0.35
SPECTRUM_TICKS = np.array([1.0e-17, 1.0e-15, 1.0e-13, 1.0e-11, 1.0e-9, 1.0e-7, 1.0e-5, 1.0e-3, 1.0e-1])

METHODS = [
    ("PHY", "N20_test_PHY-EnSRF-loc.npz", "#1F4E79"),
    ("GLA_256", "N20_test_GDLR-EnSRF-loc_global_Z256.npz", "#7B3294"),
    ("SLA_256", "N20_test_SLAT-EnSRF-loc_spatial_C2_Kz128.npz", "#2E7D32"),
    ("SLA_128", "N20_test_SLAT-EnSRF-loc_spatial_C1_Kz128.npz", "#E69F00"),
]

WINDOWS = [
    {
        "label": "Cycles 1-50",
        "analysis_cycles": "1-50",
        "analysis_sl": slice(0, 50),
        "growth_posterior_cycles": "1-49",
        "growth_next_prior_cycles": "2-50",
        "growth_post_sl": slice(0, 49),
        "growth_next_sl": slice(1, 50),
    },
    {
        "label": "Cycles 51-200",
        "analysis_cycles": "51-200",
        "analysis_sl": slice(50, 200),
        "growth_posterior_cycles": "51-199",
        "growth_next_prior_cycles": "52-200",
        "growth_post_sl": slice(50, 199),
        "growth_next_sl": slice(51, 200),
    },
]


def one_sided_power(fields: np.ndarray) -> np.ndarray:
    n = fields.shape[-1]
    fft = np.fft.rfft(fields, axis=-1)
    power = (np.abs(fft) ** 2) / float(n * n)
    if n > 1:
        power[..., 1:-1] *= 2.0
    return power


def band_sum(spec: np.ndarray, lo: int, hi: int) -> float:
    return float(np.nansum(spec[lo : hi + 1]))


def piecewise_log_power_axis(values: np.ndarray | float) -> np.ndarray:
    vals = np.asarray(values, dtype=np.float64)
    clipped = np.clip(vals, SPECTRUM_Y_MIN, SPECTRUM_Y_MAX)
    log_vals = np.log10(clipped)
    log_min = np.log10(SPECTRUM_Y_MIN)
    log_split = np.log10(SPECTRUM_Y_SPLIT)
    log_max = np.log10(SPECTRUM_Y_MAX)

    transformed = np.empty_like(log_vals)
    low_tail = log_vals <= log_split
    transformed[low_tail] = SPECTRUM_LOW_FRACTION * (log_vals[low_tail] - log_min) / (log_split - log_min)
    transformed[~low_tail] = SPECTRUM_LOW_FRACTION + (1.0 - SPECTRUM_LOW_FRACTION) * (
        log_vals[~low_tail] - log_split
    ) / (log_max - log_split)
    return transformed


def spectrum_tick_labels(ticks: np.ndarray) -> list[str]:
    return [rf"$10^{{{int(np.round(np.log10(tick)))}}}$" for tick in ticks]


def compute_window(
    truth_at_cycles: np.ndarray,
    result_arrays: dict[str, tuple[np.ndarray, np.ndarray]],
    window: dict[str, object],
    max_mode: int,
) -> tuple[dict[str, dict[str, np.ndarray]], list[dict[str, object]]]:
    analysis_sl = window["analysis_sl"]
    growth_post_sl = window["growth_post_sl"]
    growth_next_sl = window["growth_next_sl"]
    assert isinstance(analysis_sl, slice)
    assert isinstance(growth_post_sl, slice)
    assert isinstance(growth_next_sl, slice)

    analysis_truth = truth_at_cycles[analysis_sl]
    growth_post_truth = truth_at_cycles[growth_post_sl]
    growth_next_truth = truth_at_cycles[growth_next_sl]

    curves: dict[str, dict[str, np.ndarray]] = {}
    rows: list[dict[str, object]] = []

    for label, _filename, _color in METHODS:
        prior_mean_all, analysis_mean_all = result_arrays[label]

        prior_error = prior_mean_all[analysis_sl] - analysis_truth
        posterior_error = analysis_mean_all[analysis_sl] - analysis_truth
        prior_power = one_sided_power(prior_error)
        posterior_power = one_sided_power(posterior_error)
        reduction_spec = (prior_power - posterior_power).mean(axis=0)

        growth_post_error = analysis_mean_all[growth_post_sl] - growth_post_truth
        growth_next_error = prior_mean_all[growth_next_sl] - growth_next_truth
        growth_spec = (one_sided_power(growth_next_error) - one_sided_power(growth_post_error)).mean(axis=0)

        prior_spec = prior_power.mean(axis=0)
        posterior_spec = posterior_power.mean(axis=0)
        cumulative_reduction = np.cumsum(reduction_spec)
        cumulative_growth = np.cumsum(growth_spec)

        curves[label] = {
            "prior": prior_spec,
            "posterior": posterior_spec,
            "cumulative_reduction": cumulative_reduction,
            "cumulative_growth": cumulative_growth,
        }

        total_prior = band_sum(prior_spec, 1, max_mode)
        total_post = band_sum(posterior_spec, 1, max_mode)
        total_reduction = band_sum(reduction_spec, 1, max_mode)
        total_growth = band_sum(growth_spec, 1, max_mode)
        rows.append(
            {
                "window": window["label"],
                "method": label,
                "analysis_cycles": window["analysis_cycles"],
                "growth_posterior_cycles": window["growth_posterior_cycles"],
                "growth_next_prior_cycles": window["growth_next_prior_cycles"],
                "total_prior_error_power_k1_682": f"{total_prior:.10e}",
                "total_posterior_error_power_k1_682": f"{total_post:.10e}",
                "total_prior_to_posterior_reduction_k1_682": f"{total_reduction:.10e}",
                "total_posterior_to_next_prior_growth_k1_682": f"{total_growth:.10e}",
                "net_reduction_minus_growth_k1_682": f"{(total_reduction - total_growth):.10e}",
                "cumulative_reduction_k64": f"{cumulative_reduction[64]:.10e}",
                "cumulative_growth_k64": f"{cumulative_growth[64]:.10e}",
                "cumulative_reduction_k256": f"{cumulative_reduction[256]:.10e}",
                "cumulative_growth_k256": f"{cumulative_growth[256]:.10e}",
            }
        )

    return curves, rows


def main() -> None:
    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    from matplotlib.lines import Line2D

    case = np.load(TEST_CASE_PATH)
    truth_all = np.asarray(case["truth"], dtype=np.float64)
    obs_times = np.asarray(case["obs_times"], dtype=np.int64)
    truth_at_cycles = truth_all[obs_times]

    n = truth_at_cycles.shape[-1]
    max_mode = int(np.floor((n // 2) * DEALIAS_FRAC))
    modes = np.arange(n // 2 + 1)
    plot_modes = modes[1 : max_mode + 1]

    result_arrays: dict[str, tuple[np.ndarray, np.ndarray]] = {}
    for label, filename, _color in METHODS:
        with np.load(RESULT_DIR / filename) as data:
            result_arrays[label] = (
                np.asarray(data["prior_mean"], dtype=np.float64),
                np.asarray(data["analysis_mean"], dtype=np.float64),
            )

    all_curves: list[dict[str, dict[str, np.ndarray]]] = []
    rows: list[dict[str, object]] = []
    for window in WINDOWS:
        curves, window_rows = compute_window(truth_at_cycles, result_arrays, window, max_mode)
        all_curves.append(curves)
        rows.extend(window_rows)

    with OUT_SUMMARY.open("w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=list(rows[0].keys()))
        writer.writeheader()
        writer.writerows(rows)

    plt.rcParams.update(
        {
            "font.family": "DejaVu Serif",
            "font.size": 10.0,
            "mathtext.fontset": "dejavuserif",
            "axes.linewidth": 0.9,
        }
    )

    fig, axes = plt.subplots(2, 3, figsize=(14.5, 8.1), facecolor="white", sharex="col")

    for row_idx, (window, curves) in enumerate(zip(WINDOWS, all_curves)):
        row_axes = axes[row_idx]
        for label, _filename, color in METHODS:
            row_axes[0].semilogx(
                plot_modes,
                piecewise_log_power_axis(curves[label]["prior"][1 : max_mode + 1]),
                color=color,
                lw=1.55,
            )
            row_axes[1].semilogx(
                plot_modes,
                piecewise_log_power_axis(curves[label]["posterior"][1 : max_mode + 1]),
                color=color,
                lw=1.55,
            )
            row_axes[2].semilogx(
                plot_modes,
                curves[label]["cumulative_reduction"][1 : max_mode + 1],
                color=color,
                lw=1.65,
            )
            row_axes[2].semilogx(
                plot_modes,
                curves[label]["cumulative_growth"][1 : max_mode + 1],
                color=color,
                lw=1.65,
                ls="--",
        )

        row_axes[0].set_ylabel(f"{window['label']}\nError power")
        row_axes[1].set_ylabel("Error power")
        row_axes[2].set_ylabel("Cumulative error power")

    titles = ["Prior error power", "Posterior error power", "Cumulative error reduction and growth"]
    for col, title in enumerate(titles):
        axes[0, col].set_title(title, pad=7)

    tags = ["(a)", "(b)", "(c)", "(d)", "(e)", "(f)"]
    for ax, tag in zip(axes.ravel(), tags):
        ax.set_xlim(1, max_mode)
        ax.set_xlabel("")
        ax.text(
            0.025,
            0.975,
            tag,
            transform=ax.transAxes,
            ha="left",
            va="top",
            fontsize=PANEL_TAG_FONT_SIZE,
            bbox={"facecolor": "white", "edgecolor": "none", "alpha": 0.82, "pad": 1.5},
        )
        ax.grid(True, which="major", color="0.88", lw=0.8)
        ax.grid(True, which="minor", color="0.93", lw=0.5)
        ax.set_axisbelow(True)

    for ax in axes.ravel():
        ax.xaxis.label.set_size(AXIS_LABEL_FONT_SIZE)
        ax.yaxis.label.set_size(AXIS_LABEL_FONT_SIZE)

    for ax in axes[1, :]:
        ax.set_xlabel("Wavenumber")

    for ax in axes[0, :]:
        ax.tick_params(axis="x", labelbottom=False)

    spectrum_ticks = SPECTRUM_TICKS[(SPECTRUM_TICKS >= SPECTRUM_Y_MIN) & (SPECTRUM_TICKS <= SPECTRUM_Y_MAX)]
    spectrum_tick_positions = piecewise_log_power_axis(spectrum_ticks)
    for ax in list(axes[:, 0]) + list(axes[:, 1]):
        ax.set_ylim(float(piecewise_log_power_axis(SPECTRUM_Y_MIN)), float(piecewise_log_power_axis(SPECTRUM_Y_MAX)))
        ax.set_yticks(spectrum_tick_positions)
        ax.set_yticklabels(spectrum_tick_labels(spectrum_ticks))
        ax.axhline(float(piecewise_log_power_axis(SPECTRUM_Y_SPLIT)), color="0.78", lw=0.65, ls=":")

    for ax in axes[:, 2]:
        ax.axhline(0.0, color="0.2", lw=0.85)
        ax.axvline(64, color="0.6", lw=0.85, ls=":")
        ax.axvline(256, color="0.6", lw=0.85, ls=":")

    method_handles = [Line2D([0], [0], color=color, lw=2.5, label=label) for label, _filename, color in METHODS]
    style_handles = [
        Line2D([0], [0], color="0.2", lw=1.8, label="analysis error reduction"),
        Line2D([0], [0], color="0.2", lw=1.8, ls="--", label="forecast error growth"),
    ]
    fig.legend(
        handles=method_handles + style_handles,
        loc="upper center",
        ncol=6,
        frameon=False,
        bbox_to_anchor=(0.54, 0.975),
        columnspacing=1.25,
        handlelength=2.0,
    )
    fig.tight_layout(rect=(0.02, 0.02, 0.995, 0.905))
    fig.savefig(OUT_BASE.with_suffix(".png"), dpi=300)
    fig.savefig(OUT_BASE.with_suffix(".pdf"))
    plt.close(fig)

    print(OUT_BASE.with_suffix(".png"))
    print(OUT_BASE.with_suffix(".pdf"))
    print(OUT_SUMMARY)


if __name__ == "__main__":
    main()



from __future__ import annotations

import argparse
from pathlib import Path

import numpy as np

from example_common import make_params
from lda.ks import integrate_steps, random_initial_state
from lda.utils import ensure_dir


def main() -> None:
    parser = argparse.ArgumentParser(description="Generate a dispersive-KS truth trajectory.")
    parser.add_argument("--n", type=int, default=512)
    parser.add_argument("--length", type=float, default=None)
    parser.add_argument("--dt", type=float, default=0.05)
    parser.add_argument("--spinup-steps", type=int, default=2000)
    parser.add_argument("--save-steps", type=int, default=800)
    parser.add_argument("--save-every", type=int, default=20)
    parser.add_argument("--seed", type=int, default=1)
    parser.add_argument("--output", type=Path, default=Path("example_outputs/truth_run.npz"))
    args = parser.parse_args()

    params = make_params(args.n, args.length)
    rng = np.random.default_rng(args.seed)
    state = random_initial_state(params, rng, std=0.5, max_initial_mode=min(64, params.n // 4))
    state = integrate_steps(state, args.spinup_steps, args.dt, params)

    states = [state.copy()]
    times = [args.spinup_steps * args.dt]
    for step in range(args.save_every, args.save_steps + 1, args.save_every):
        state = integrate_steps(state, args.save_every, args.dt, params)
        states.append(state.copy())
        times.append((args.spinup_steps + step) * args.dt)

    ensure_dir(args.output.parent)
    np.savez_compressed(
        args.output,
        truth=np.asarray(states),
        time=np.asarray(times),
        n=args.n,
        length=params.length,
        dt=args.dt,
        delta3=params.third_order_dispersion,
        delta5=params.fifth_order_dispersion,
    )
    print(f"saved {args.output}")


if __name__ == "__main__":
    main()



from __future__ import annotations

import sys
from pathlib import Path

import numpy as np
import torch

ROOT = Path(__file__).resolve().parents[1]
CORE = ROOT / "core"
if str(CORE) not in sys.path:
    sys.path.insert(0, str(CORE))

from lda.ks import KSParams, integrate_steps, random_initial_state  # noqa: E402
from lda.utils import cyclic_distances, ensemble_spread, gaspari_cohn, rmse  # noqa: E402


def make_params(
    n: int,
    length: float | None,
    delta3: float = 0.1,
    delta5: float = 0.1,
) -> KSParams:
    if length is None:
        length = 0.125 * n
    return KSParams(
        n=n,
        length=length,
        linear_growth=1.0,
        third_order_dispersion=delta3,
        fifth_order_dispersion=delta5,
        fourth_order_damping=1.0,
    )


def forecast_ensemble(
    ensemble: np.ndarray,
    n_steps: int,
    dt: float,
    params: KSParams,
) -> np.ndarray:
    return np.stack([integrate_steps(member, n_steps, dt, params) for member in ensemble])


def make_truth(params: KSParams, rng: np.random.Generator, spinup_steps: int, dt: float) -> np.ndarray:
    state = random_initial_state(params, rng, std=0.5, max_initial_mode=min(64, params.n // 4))
    return integrate_steps(state, spinup_steps, dt, params)


def make_initial_ensemble(
    truth: np.ndarray,
    params: KSParams,
    rng: np.random.Generator,
    n_members: int,
    perturbation_std: float,
) -> np.ndarray:
    members = []
    for _ in range(n_members):
        perturb = random_initial_state(
            params,
            rng,
            std=perturbation_std,
            max_initial_mode=min(64, params.n // 4),
        )
        members.append(truth + perturb)
    return np.asarray(members)


def inflate_ensemble(ensemble: np.ndarray, factor: float) -> np.ndarray:
    mean = ensemble.mean(axis=0, keepdims=True)
    return mean + factor * (ensemble - mean)


def physical_localization(obs_loc: int, n: int, radius: float) -> np.ndarray:
    distance = cyclic_distances(np.arange(n), float(obs_loc), n)
    return gaspari_cohn(distance, radius)


def spatial_latent_localization(
    obs_loc: int,
    n: int,
    latent_channels: int,
    kz: int,
    radius: float,
) -> np.ndarray:
    latent_spacing = n / float(kz)
    latent_positions = np.arange(kz) * latent_spacing
    distance = cyclic_distances(latent_positions, float(obs_loc), n)
    weights = gaspari_cohn(distance, radius)
    return np.tile(weights, latent_channels)


def ensrf_scalar_update(
    ensemble: np.ndarray,
    y_ens: np.ndarray,
    obs_value: float,
    obs_error_var: float,
    loc_weights: np.ndarray | None = None,
) -> np.ndarray:
    n_members = ensemble.shape[0]
    x_mean = ensemble.mean(axis=0)
    x_pert = ensemble - x_mean
    y_mean = float(y_ens.mean())
    y_pert = y_ens - y_mean
    var_y = float(y_pert @ y_pert / max(n_members - 1, 1))
    denom = var_y + obs_error_var
    if not np.isfinite(denom) or denom <= 0.0:
        return ensemble

    cov_xy = y_pert[:, None].T @ x_pert / max(n_members - 1, 1)
    cov_xy = cov_xy.ravel()
    if loc_weights is not None:
        cov_xy = cov_xy * loc_weights

    gain = cov_xy / denom
    alpha = 1.0 / (1.0 + np.sqrt(obs_error_var / denom))
    new_mean = x_mean + gain * (obs_value - y_mean)
    new_pert = x_pert - alpha * np.outer(y_pert, gain)
    return new_mean + new_pert


def physical_ensrf(
    ensemble: np.ndarray,
    obs_values: np.ndarray,
    obs_locs: np.ndarray,
    obs_error_std: float,
    loc_radius: float | None,
) -> np.ndarray:
    analysis = ensemble.copy()
    obs_error_var = obs_error_std**2
    for obs_value, obs_loc in zip(obs_values, obs_locs):
        y_ens = analysis[:, obs_loc]
        loc = None if loc_radius is None else physical_localization(int(obs_loc), analysis.shape[1], loc_radius)
        analysis = ensrf_scalar_update(analysis, y_ens, float(obs_value), obs_error_var, loc)
    return analysis


def load_checkpoint(model: torch.nn.Module, checkpoint: str | None) -> torch.nn.Module:
    if checkpoint:
        state = torch.load(checkpoint, map_location="cpu")
        state_dict = state.get("model_state_dict", state.get("state_dict", state))
        model.load_state_dict(state_dict, strict=False)
    model.eval()
    return model


def encode_states(model: torch.nn.Module, states: np.ndarray, mean: float, std: float) -> np.ndarray:
    x = torch.as_tensor(((states - mean) / std)[:, None, :], dtype=torch.float32)
    with torch.no_grad():
        z = model.encode(x).detach().cpu().numpy()
    return z


def decode_states(model: torch.nn.Module, latent: np.ndarray, mean: float, std: float) -> np.ndarray:
    z = torch.as_tensor(latent, dtype=torch.float32)
    with torch.no_grad():
        x = model.decode(z).detach().cpu().numpy()[:, 0, :]
    return x * std + mean


def global_latent_localized_ensrf(
    ensemble: np.ndarray,
    obs_values: np.ndarray,
    obs_locs: np.ndarray,
    obs_error_std: float,
    loc_radius: float,
    model: torch.nn.Module,
    norm_mean: float,
    norm_std: float,
) -> np.ndarray:
    physical = ensemble.copy()
    obs_error_var = obs_error_std**2
    for obs_value, obs_loc in zip(obs_values, obs_locs):
        z = encode_states(model, physical, norm_mean, norm_std)
        decoded = decode_states(model, z, norm_mean, norm_std)
        y_ens = decoded[:, obs_loc]
        z_flat = z.reshape(z.shape[0], -1)
        z_flat_a = ensrf_scalar_update(z_flat, y_ens, float(obs_value), obs_error_var, None)
        decoded_a = decode_states(model, z_flat_a, norm_mean, norm_std)
        loc = physical_localization(int(obs_loc), physical.shape[1], loc_radius)
        physical = physical + (decoded_a - physical) * loc[None, :]
    return physical


def spatial_latent_ensrf(
    ensemble: np.ndarray,
    obs_values: np.ndarray,
    obs_locs: np.ndarray,
    obs_error_std: float,
    loc_radius: float,
    model: torch.nn.Module,
    norm_mean: float,
    norm_std: float,
) -> np.ndarray:
    z = encode_states(model, ensemble, norm_mean, norm_std)
    n_members, latent_channels, kz = z.shape
    z_flat = z.reshape(n_members, -1)
    obs_error_var = obs_error_std**2
    for obs_value, obs_loc in zip(obs_values, obs_locs):
        decoded = decode_states(model, z_flat.reshape(n_members, latent_channels, kz), norm_mean, norm_std)
        y_ens = decoded[:, obs_loc]
        loc = spatial_latent_localization(int(obs_loc), ensemble.shape[1], latent_channels, kz, loc_radius)
        z_flat = ensrf_scalar_update(z_flat, y_ens, float(obs_value), obs_error_var, loc)
    return decode_states(model, z_flat.reshape(n_members, latent_channels, kz), norm_mean, norm_std)


def diagnostics(prior: np.ndarray, analysis: np.ndarray, truth: np.ndarray) -> tuple[float, float, float, float]:
    prior_mean = prior.mean(axis=0)
    analysis_mean = analysis.mean(axis=0)
    return (
        rmse(prior_mean, truth),
        rmse(analysis_mean, truth),
        ensemble_spread(prior),
        ensemble_spread(analysis),
    )



from __future__ import annotations

from pathlib import Path
import csv
import math
import os

import numpy as np


SCRIPT_DIR = Path(__file__).resolve().parent
ROOT = SCRIPT_DIR.parent
MPLCONFIGDIR = ROOT / ".mplconfig_pic_fig04_sawtooth_n20"
MPLCONFIGDIR.mkdir(exist_ok=True)
for lock_file in MPLCONFIGDIR.glob("*.matplotlib-lock"):
    try:
        lock_file.unlink()
    except OSError:
        pass
os.environ.setdefault("MPLCONFIGDIR", str(MPLCONFIGDIR))

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D


NOINFL_DIR = (
    ROOT
    / "outputs"
    / "PATH_NAME"
    / "results"
    / "four_convae_testoracle_unlocinf100_physharedloc"
)
PHYSHARED_DIR = (
    ROOT
    / "outputs"
    / "PATH_NAME"
    / "results"
    / "phyloc_inflation_shared"
)

OUT_PNG = SCRIPT_DIR / "Fig04_physhared_N20_sawtooth_ladder.png"
OUT_PDF = SCRIPT_DIR / "Fig04_physhared_N20_sawtooth_ladder.pdf"
OUT_CSV = SCRIPT_DIR / "Fig04_physhared_N20_sawtooth_ladder_sources.csv"
TEST_CASE_PATH = (
    ROOT
    / "outputs"
    / "PATH_NAME"
    / "data"
    / "test_case.npz"
)
RMSE_SCALE = float(np.std(np.load(TEST_CASE_PATH)["truth"]))

N_CYCLES = 200
DISCARD_CYCLES = 50

METHODS = [
    ("PHY", "#1F4E79"),
    ("GLA_256", "#7B3294"),
    ("GLA_128", "#D81B60"),
    ("SLA_256", "#2E7D32"),
    ("SLA_128", "#E69F00"),
]
NO_DA_COLOR = "0.0"
LEFT_LINEWIDTH = 1.15

LEFT_ROWS = [
    (
        "No localization, no inflation",
        NOINFL_DIR,
        {
            "No DA": "test_NODA.npz",
            "PHY": "test_PHY-EnSRF-unloc.npz",
            "GLA_256": "test_GLAT-EnSRF-unloc_global_Z256.npz",
            "GLA_128": "test_GLAT-EnSRF-unloc_global_Z128.npz",
            "SLA_256": "test_LAT-EnSRF-unloc_spatial_C2_Kz128.npz",
            "SLA_128": "test_LAT-EnSRF-unloc_spatial_C1_Kz128.npz",
        },
    ),
    (
        "Inflation only",
        PHYSHARED_DIR,
        {
            "No DA": "N20_test_NODA.npz",
            "PHY": "N20_test_PHY-EnSRF-unloc.npz",
            "GLA_256": "N20_test_GLAT-EnSRF-unloc_global_Z256.npz",
            "GLA_128": "N20_test_GLAT-EnSRF-unloc_global_Z128.npz",
            "SLA_256": "N20_test_LAT-EnSRF-unloc_spatial_C2_Kz128.npz",
            "SLA_128": "N20_test_LAT-EnSRF-unloc_spatial_C1_Kz128.npz",
        },
    ),
    (
        "Localization + inflation",
        PHYSHARED_DIR,
        {
            "No DA": "N20_test_NODA.npz",
            "PHY": "N20_test_PHY-EnSRF-loc.npz",
            "GLA_256": "N20_test_GDLR-EnSRF-loc_global_Z256.npz",
            "GLA_128": "N20_test_GDLR-EnSRF-loc_global_Z128.npz",
            "SLA_256": "N20_test_SLAT-EnSRF-loc_spatial_C2_Kz128.npz",
            "SLA_128": "N20_test_SLAT-EnSRF-loc_spatial_C1_Kz128.npz",
        },
    ),
]

RIGHT_DIFFS = [
    ("GLA_256 - PHY", "GLA_256", "#7B3294"),
    ("SLA_256 - PHY", "SLA_256", "#2E7D32"),
    ("SLA_128 - PHY", "SLA_128", "#E69F00"),
]


def load_curve(base_dir: Path, filename: str) -> dict[str, np.ndarray | str | bool | int | float]:
    path = base_dir / filename
    if not path.exists():
        raise FileNotFoundError(path)
    data = np.load(path)
    prior = data["prior_nrmse"].astype(float) * RMSE_SCALE
    analysis = data["analysis_nrmse"].astype(float) * RMSE_SCALE
    finite = np.isfinite(prior) & np.isfinite(analysis)
    first_bad = int(np.argmax(~finite)) if not np.all(finite) else -1
    return {
        "path": str(path),
        "prior": prior,
        "analysis": analysis,
        "stable": bool(np.all(finite)),
        "first_bad": first_bad,
        "mean_prior": float(np.nanmean(prior[DISCARD_CYCLES:])),
        "mean_analysis": float(np.nanmean(analysis[DISCARD_CYCLES:])),
    }


def sawtooth_xy(prior: np.ndarray, analysis: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    x = np.empty(prior.size * 2, dtype=float)
    y = np.empty(prior.size * 2, dtype=float)
    cycles = np.arange(1, prior.size + 1, dtype=float)
    x[0::2] = cycles - 0.35
    x[1::2] = cycles + 0.35
    y[0::2] = prior
    y[1::2] = analysis
    return x, y


def write_source_table(loaded: dict[str, dict[str, dict[str, np.ndarray | str | bool | int | float]]]) -> None:
    fields = [
        "setting",
        "method_label",
        "source_npz",
        "normalization_truth_std",
        "stable",
        "first_nonfinite_cycle_1based",
        "mean_prior_rmse_after_discard",
        "mean_analysis_rmse_after_discard",
    ]
    with OUT_CSV.open("w", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=fields)
        writer.writeheader()
        for setting, curves in loaded.items():
            for label, curve in curves.items():
                first_bad = int(curve["first_bad"])
                mean_prior = float(curve["mean_prior"])
                mean_analysis = float(curve["mean_analysis"])
                writer.writerow(
                    {
                        "setting": setting,
                        "method_label": label,
                        "source_npz": curve["path"],
                        "normalization_truth_std": f"{RMSE_SCALE:.8f}",
                        "stable": curve["stable"],
                        "first_nonfinite_cycle_1based": "" if first_bad < 0 else first_bad + 1,
                        "mean_prior_rmse_after_discard": f"{mean_prior:.8f}"
                        if math.isfinite(mean_prior)
                        else "nan",
                        "mean_analysis_rmse_after_discard": f"{mean_analysis:.8f}"
                        if math.isfinite(mean_analysis)
                        else "nan",
                    }
                )


def main() -> None:
    loaded: dict[str, dict[str, dict[str, np.ndarray | str | bool | int | float]]] = {}
    for setting, base_dir, files in LEFT_ROWS:
        loaded[setting] = {label: load_curve(base_dir, filename) for label, filename in files.items()}
    write_source_table(loaded)

    plt.rcParams.update(
        {
            "font.family": "DejaVu Serif",
            "font.size": 10.5,
            "mathtext.fontset": "dejavuserif",
            "axes.linewidth": 0.9,
        }
    )

    fig, axes = plt.subplots(
        3,
        2,
        figsize=(12.3, 8.8),
        sharex="col",
        gridspec_kw={"width_ratios": [1.45, 1.0], "wspace": 0.18, "hspace": 0.20},
        facecolor="white",
    )

    left_panel_labels = ["(a)", "(b)", "(c)"]
    right_panel_labels = ["(d)", "(e)", "(f)"]

    left_ylims = [
        (1.00, 2.30),
        (1.00, 2.30),
        (0.10, 1.70),
    ]
    right_ylim = (-0.55 * RMSE_SCALE, 0.58 * RMSE_SCALE)

    for row_index, (setting, _, _) in enumerate(LEFT_ROWS):
        ax = axes[row_index, 0]
        ax.axvspan(0.5, DISCARD_CYCLES + 0.5, color="0.92", lw=0, zorder=0)
        for label, color in METHODS:
            curve = loaded[setting][label]
            x, y = sawtooth_xy(curve["prior"], curve["analysis"])
            ax.plot(x, y, color=color, lw=LEFT_LINEWIDTH, alpha=0.88, label=label)
            if not curve["stable"]:
                first_bad = int(curve["first_bad"])
                if first_bad > 0:
                    ax.plot(
                        first_bad + 0.5,
                        float(curve["analysis"][first_bad - 1]),
                        marker="x",
                        color=color,
                        ms=7,
                        mew=1.6,
                    )
                    ax.text(
                        first_bad + 4.0,
                        float(curve["analysis"][first_bad - 1]) + 0.04,
                        "unstable",
                        color=color,
                        fontsize=8.5,
                    )
        ax.text(0.012, 0.93, left_panel_labels[row_index], transform=ax.transAxes, ha="left", va="top")
        if row_index < 3:
            noda_curve = loaded[setting]["No DA"]
            ax.plot(
                np.arange(1, N_CYCLES + 1, dtype=float),
                noda_curve["analysis"],
                color=NO_DA_COLOR,
                lw=LEFT_LINEWIDTH,
                ls=(0, (4.0, 2.2)),
                alpha=0.85,
                label="No DA",
            )
        ax.set_ylabel("RMSE")
        ax.set_xlim(0.5, N_CYCLES + 0.5)
        ax.set_ylim(*left_ylims[row_index])
        ax.grid(True, axis="y", color="0.88", lw=0.8)
        ax.set_axisbelow(True)
        if row_index == 0:
            legend_handles = [
                Line2D([0], [0], color=NO_DA_COLOR, lw=LEFT_LINEWIDTH, ls=(0, (4.0, 2.2)), label="NoDA"),
                Line2D([0], [0], color="#1F4E79", lw=LEFT_LINEWIDTH, label="PHY"),
                Line2D([0], [0], color="#7B3294", lw=LEFT_LINEWIDTH, label="GLA_256"),
                Line2D([0], [0], color="#D81B60", lw=LEFT_LINEWIDTH, label="GLA_128"),
                Line2D([0], [0], color="#2E7D32", lw=LEFT_LINEWIDTH, label="SLA_256"),
                Line2D([0], [0], color="#E69F00", lw=LEFT_LINEWIDTH, label="SLA_128"),
            ]
            ax.legend(
                handles=legend_handles,
                loc="upper center",
                bbox_to_anchor=(0.58, 1.28),
                ncol=6,
                frameon=False,
                handlelength=2.0,
                columnspacing=0.85,
            )

    loc_setting = "Localization + inflation"
    phy_analysis = loaded[loc_setting]["PHY"]["analysis"]
    cycles = np.arange(1, N_CYCLES + 1, dtype=float)
    right_text_layouts = [
        {
            "title_xy": (0.03, 0.08),
            "title_ha": "left",
            "title_va": "bottom",
            "mean_xy": (0.98, 0.08),
            "mean_ha": "right",
            "mean_va": "bottom",
        },
        {
            "title_xy": (0.03, 0.08),
            "title_ha": "left",
            "title_va": "bottom",
            "mean_xy": (0.98, 0.08),
            "mean_ha": "right",
            "mean_va": "bottom",
        },
        {
            "title_xy": (0.03, 0.08),
            "title_ha": "left",
            "title_va": "bottom",
            "mean_xy": (0.98, 0.08),
            "mean_ha": "right",
            "mean_va": "bottom",
        },
    ]
    for row_index, (title, label, color) in enumerate(RIGHT_DIFFS):
        ax = axes[row_index, 1]
        layout = right_text_layouts[row_index]
        ax.axhspan(right_ylim[0], 0.0, color="#EAF4EA", lw=0, zorder=-2)
        ax.axhspan(0.0, right_ylim[1], color="#F8ECEC", lw=0, zorder=-2)
        ax.axvspan(0.5, DISCARD_CYCLES + 0.5, color="0.92", lw=0, zorder=0)
        curve = loaded[loc_setting][label]
        diff = curve["analysis"] - phy_analysis
        ax.axhline(0.0, color="0.25", lw=0.9)
        ax.plot(cycles, diff, color=color, lw=1.35)
        ax.text(0.02, 0.93, right_panel_labels[row_index], transform=ax.transAxes, ha="left", va="top")
        ax.text(
            layout["title_xy"][0],
            layout["title_xy"][1],
            title,
            transform=ax.transAxes,
            ha=layout["title_ha"],
            va=layout["title_va"],
            fontsize=9.0,
            bbox=dict(facecolor="white", edgecolor="none", alpha=0.92, pad=1.8),
        )
        mean_diff = float(np.nanmean(diff[DISCARD_CYCLES:]))
        ax.text(
            layout["mean_xy"][0],
            layout["mean_xy"][1],
            f"mean = {mean_diff:+.3f}",
            transform=ax.transAxes,
            ha=layout["mean_ha"],
            va=layout["mean_va"],
            fontsize=9.0,
            color=color,
            bbox=dict(facecolor="white", edgecolor="none", alpha=0.88, pad=1.6),
        )
        ax.set_ylabel("Analysis RMSE difference")
        ax.set_xlim(0.5, N_CYCLES + 0.5)
        ax.set_ylim(*right_ylim)
        ax.grid(True, axis="y", color="0.88", lw=0.8)
        ax.set_axisbelow(True)

    for ax in axes[-1, :]:
        ax.set_xlabel("DA cycle")

    for ax in axes.ravel():
        ax.tick_params(direction="out", length=4, width=0.8)

    fig.savefig(OUT_PNG, dpi=300, bbox_inches="tight")
    fig.savefig(OUT_PDF, bbox_inches="tight")
    print(f"saved {OUT_PNG}")
    print(f"saved {OUT_PDF}")
    print(f"saved {OUT_CSV}")


if __name__ == "__main__":
    main()



from __future__ import annotations

import csv
import os
import shutil
from pathlib import Path

import numpy as np


ROOT = Path(r"D:\LDA")
PIC = ROOT / "PIC"
PRIOR_DIR = (
    ROOT
    / "outputs"
    / "PATH_NAME"
    / "results"
    / "prior_ensembles"
)

ORG_FIG_PATH = PIC / "Organized" / "Fig09_abs_N10_N40_corr_difference_early_late_full_PHY_CBR_wet_vmax055.png"
ORG_FIG_PDF = PIC / "Organized" / "Fig09_abs_N10_N40_corr_difference_early_late_full_PHY_CBR_wet_vmax055.pdf"
SUMMARY_CSV = PIC / "Fig09_abs_N10_N40_corr_difference_early_late_full_PHY_summary.csv"
CBR_WET = Path(r"CBR_wet.rgb")

PHYSICAL_LOC_RADIUS = 192
LATENT_GRID_SIZE = 128
LATENT_GRID_SPACING = 2048 // LATENT_GRID_SIZE
LATENT_LOC_RADIUS = PHYSICAL_LOC_RADIUS // LATENT_GRID_SPACING
COLOR_LIMIT = 0.55

N10_CASE = "N10_seed"
N40_CASE = "N40_seed"

WINDOWS = [
    ("Cycles 1-50", slice(0, 50)),
    ("Cycles 51-200", slice(50, 200)),
]

SPACES = [
    ("PHY", "physical", "PHY"),
    ("GLA_256", "global_latent", "GLA_256"),
    ("SLA_256", "stored_latent", "SLA_256"),
    ("SLA_128", "stored_latent", "SLA_128"),
]


def load_prior_npz(case_key: str, method: str) -> dict[str, np.ndarray]:
    path = PRIOR_DIR / f"{case_key}_{method}_physhared_prior_ensembles.npz"
    if not path.exists():
        raise FileNotFoundError(path)
    with np.load(path) as data:
        return {key: data[key] for key in data.files}


def all_cycle_variables(data: dict[str, np.ndarray], space_kind: str) -> np.ndarray:
    if space_kind == "physical":
        return data["prior_ensemble"]
    if space_kind in {"stored_latent", "global_latent"}:
        latent = data["prior_latent_ensemble"]
        return latent.reshape(latent.shape[0], latent.shape[1], -1)
    raise ValueError(space_kind)


def corr_matrix(variables: np.ndarray) -> np.ndarray:
    x = np.asarray(variables, dtype=np.float64)
    x = x - x.mean(axis=0, keepdims=True)
    std = x.std(axis=0, ddof=1)
    cov = x.T @ x / float(x.shape[0] - 1)
    denom = std[:, None] * std[None, :]
    corr = np.divide(cov, denom, out=np.zeros_like(cov), where=denom > 1.0e-14)
    return np.clip(corr, -1.0, 1.0)


def mean_abs_difference(
    n10_vars: np.ndarray,
    n40_vars: np.ndarray,
    window: slice,
) -> tuple[np.ndarray, int]:
    n_cycles = min(n10_vars.shape[0], n40_vars.shape[0])
    start = int(window.start or 0)
    stop = int(window.stop if window.stop is not None else n_cycles)
    stop = min(stop, n_cycles)
    abs_sum: np.ndarray | None = None
    count = 0
    for cycle in range(start, stop):
        diff_abs = np.abs(corr_matrix(n10_vars[cycle]) - corr_matrix(n40_vars[cycle]))
        if abs_sum is None:
            abs_sum = np.zeros_like(diff_abs)
        abs_sum += diff_abs
        count += 1
    if abs_sum is None or count == 0:
        raise RuntimeError("No paired cycles were available.")
    return abs_sum / count, count


def offdiag_mask(n: int) -> np.ndarray:
    return ~np.eye(n, dtype=bool)


def periodic_band_mask(n: int, radius: int) -> np.ndarray:
    idx = np.arange(n)
    d = np.abs(idx[:, None] - idx[None, :])
    d = np.minimum(d, n - d)
    return (d <= radius) & offdiag_mask(n)


def sla256_localization_mask() -> np.ndarray:
    n = 2 * LATENT_GRID_SIZE
    cell = np.arange(n) % LATENT_GRID_SIZE
    d = np.abs(cell[:, None] - cell[None, :])
    d = np.minimum(d, LATENT_GRID_SIZE - d)
    return (d <= LATENT_LOC_RADIUS) & offdiag_mask(n)


def localization_mask(label: str, n: int) -> np.ndarray | None:
    if label == "PHY":
        return periodic_band_mask(n, PHYSICAL_LOC_RADIUS)
    if label == "SLA_128":
        return periodic_band_mask(n, LATENT_LOC_RADIUS)
    if label == "SLA_256":
        return sla256_localization_mask()
    return None


def draw_periodic_band_block(ax, x_start: int, y_start: int, size: int, radius: int) -> None:
    x_end = x_start + size - 1
    y_end = y_start + size - 1
    r = int(radius)
    if r <= 0 or r >= size:
        return
    style = {
        "color": "0.0",
        "lw": 1.75,
        "ls": (0, (4, 3)),
        "alpha": 0.78,
        "solid_capstyle": "butt",
        "zorder": 4,
    }
    ax.plot([x_start, x_end - r], [y_start + r, y_end], **style)
    ax.plot([x_start + r, x_end], [y_start, y_end - r], **style)
    ax.plot([x_start, x_start + r], [y_end - r, y_end], **style)
    ax.plot([x_end - r, x_end], [y_start, y_start + r], **style)


def add_localization_guides(ax, label: str) -> None:
    if label == "PHY":
        draw_periodic_band_block(ax, 0, 0, 2048, PHYSICAL_LOC_RADIUS)
    elif label == "SLA_128":
        draw_periodic_band_block(ax, 0, 0, LATENT_GRID_SIZE, LATENT_LOC_RADIUS)
    elif label == "SLA_256":
        for x0 in (0, LATENT_GRID_SIZE):
            for y0 in (0, LATENT_GRID_SIZE):
                draw_periodic_band_block(ax, x0, y0, LATENT_GRID_SIZE, LATENT_LOC_RADIUS)


def add_sla256_block_guides(ax) -> None:
    boundary = LATENT_GRID_SIZE - 0.5
    style = {
        "color": "0.28",
        "lw": 2.05,
        "ls": (0, (3, 3)),
        "alpha": 0.9,
        "zorder": 5,
    }
    ax.axvline(boundary, **style)
    ax.axhline(boundary, **style)


def load_cbr_wet_cmap():
    import matplotlib.colors as mcolors

    if not CBR_WET.exists():
        return "magma_r"
    colors = []
    with CBR_WET.open("r", encoding="utf-8") as f:
        for line in f:
            parts = line.split()
            if len(parts) < 3:
                continue
            try:
                rgb = [float(parts[0]), float(parts[1]), float(parts[2])]
            except ValueError:
                continue
            colors.append(rgb)
    arr = np.asarray(colors, dtype=float)
    if arr.size == 0:
        return "magma_r"
    if np.nanmax(arr) > 1.0:
        arr = arr / 255.0
    return mcolors.LinearSegmentedColormap.from_list("CBR_wet", arr)


def set_axis_labels(ax, label: str, row: int) -> None:
    if label == "PHY":
        ticks = np.array([0, 512, 1024, 1536, 2047])
        ax.set_xticks(ticks)
        ax.set_yticks(ticks)
        ax.set_yticklabels([str(int(t)) for t in ticks])
        if row == 1:
            ax.set_xticklabels([str(int(t)) for t in ticks])
            ax.set_xlabel("Physical grid index")
        else:
            ax.set_xticklabels([])
        ax.set_ylabel("Physical grid index")
    else:
        n = 128 if label == "SLA_128" else 256
        step = 25 if n == 128 else 50
        ticks = np.arange(0, n, step)
        ax.set_xticks(ticks)
        ax.set_yticks(ticks)
        ax.set_yticklabels([str(int(t)) for t in ticks])
        if row == 1:
            ax.set_xticklabels([str(int(t)) for t in ticks])
            ax.set_xlabel("Latent variable index")
        else:
            ax.set_xticklabels([])
        ax.set_ylabel("Latent variable index")


def main() -> None:
    mpl_config = ROOT / ".mplconfig_fig34_abscorr"
    mpl_config.mkdir(exist_ok=True)
    os.environ["MPLCONFIGDIR"] = str(mpl_config)

    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt

    loaded: dict[tuple[str, str], np.ndarray] = {}
    for label, space_kind, method in SPACES:
        for case_key in (N10_CASE, N40_CASE):
            data = load_prior_npz(case_key, method)
            loaded[(case_key, label)] = all_cycle_variables(data, space_kind)

    maps: dict[tuple[str, str], np.ndarray] = {}
    rows: list[dict[str, object]] = []
    for window_label, window_slice in WINDOWS:
        for label, _space_kind, _method in SPACES:
            abs_diff, count = mean_abs_difference(
                loaded[(N10_CASE, label)],
                loaded[(N40_CASE, label)],
                window_slice,
            )
            maps[(window_label, label)] = abs_diff

            offdiag = offdiag_mask(abs_diff.shape[0])
            loc = localization_mask(label, abs_diff.shape[0])
            row = {
                "window": window_label,
                "label": label,
                "n_variables": int(abs_diff.shape[0]),
                "paired_cycles": int(count),
                "mean_abs_offdiag_cycle_abs_N10_minus_N40": f"{float(np.mean(abs_diff[offdiag])):.10e}",
                "median_abs_offdiag_cycle_abs_N10_minus_N40": f"{float(np.median(abs_diff[offdiag])):.10e}",
                "p90_abs_offdiag_cycle_abs_N10_minus_N40": f"{float(np.percentile(abs_diff[offdiag], 90)):.10e}",
                "mean_abs_localized_cycle_abs_N10_minus_N40": "",
                "median_abs_localized_cycle_abs_N10_minus_N40": "",
                "p90_abs_localized_cycle_abs_N10_minus_N40": "",
            }
            if loc is not None:
                row["mean_abs_localized_cycle_abs_N10_minus_N40"] = f"{float(np.mean(abs_diff[loc])):.10e}"
                row["median_abs_localized_cycle_abs_N10_minus_N40"] = f"{float(np.median(abs_diff[loc])):.10e}"
                row["p90_abs_localized_cycle_abs_N10_minus_N40"] = f"{float(np.percentile(abs_diff[loc], 90)):.10e}"
            rows.append(row)

    with SUMMARY_CSV.open("w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=list(rows[0].keys()))
        writer.writeheader()
        writer.writerows(rows)

    matplotlib.rcParams.update(
        {
            "font.family": "serif",
            "font.size": 11.5,
            "axes.linewidth": 1.0,
            "xtick.direction": "out",
            "ytick.direction": "out",
        }
    )

    fig, axes = plt.subplots(2, 4, figsize=(16.8, 6.4), constrained_layout=False)
    panel_letters = list("abcdefgh")
    cmap = load_cbr_wet_cmap()
    for row_idx, (window_label, _window_slice) in enumerate(WINDOWS):
        for col_idx, (label, _kind, _method) in enumerate(SPACES):
            ax = axes[row_idx, col_idx]
            data = maps[(window_label, label)]
            im = ax.imshow(
                data,
                origin="lower",
                cmap=cmap,
                vmin=0.0,
                vmax=COLOR_LIMIT,
                interpolation="nearest",
                aspect="auto",
            )
            add_localization_guides(ax, label)
            if label == "SLA_256":
                add_sla256_block_guides(ax)
            set_axis_labels(ax, label, row_idx)
            letter = panel_letters[row_idx * 4 + col_idx]
            ax.text(
                0.02,
                0.96,
                f"({letter}) {label}",
                transform=ax.transAxes,
                ha="left",
                va="top",
                fontsize=13.0,
                bbox={"facecolor": "white", "alpha": 0.78, "edgecolor": "none", "pad": 1.5},
            )
    fig.subplots_adjust(left=0.105, right=0.91, bottom=0.105, top=0.96, wspace=0.36, hspace=0.20)
    for row_idx, (window_label, _window_slice) in enumerate(WINDOWS):
        pos = axes[row_idx, 0].get_position()
        fig.text(
            0.052,
            0.5 * (pos.y0 + pos.y1),
            window_label,
            rotation=90,
            ha="center",
            va="center",
            fontsize=13.0,
            family="serif",
        )
    cax = fig.add_axes([0.925, 0.19, 0.018, 0.62])
    cb = fig.colorbar(im, cax=cax, extend="max")
    cb.set_label(r"Mean $|r_{10}-r_{40}|$", rotation=90, labelpad=17)

    fig.savefig(FIG_PATH, dpi=300)
    fig.savefig(FIG_PDF)
    ORG_FIG_PATH.parent.mkdir(exist_ok=True)
    shutil.copy2(FIG_PATH, ORG_FIG_PATH)
    shutil.copy2(FIG_PDF, ORG_FIG_PDF)
    print(f"Saved {FIG_PATH}")
    print(f"Saved {ORG_FIG_PATH}")
    print(f"Saved {SUMMARY_CSV}")


if __name__ == "__main__":
    main()



from __future__ import annotations

import csv
import os
import sys
from pathlib import Path
from typing import Any

import numpy as np
import yaml


ROOT = Path(r"D:\LDA")
PIC = ROOT / "PIC"
MPLCONFIGDIR = ROOT / ".mplconfig_pic_fig08_projection_increment"
MPLCONFIGDIR.mkdir(exist_ok=True)
for lock_file in MPLCONFIGDIR.glob("*.matplotlib-lock"):
    try:
        lock_file.unlink()
    except OSError:
        pass
os.environ["MPLCONFIGDIR"] = str(MPLCONFIGDIR)

sys.path.insert(0, str(ROOT))

from lda.ks_ae import (  # noqa: E402
    checkpoint_path,
    decode_global_latent,
    decode_latent,
    encode_global_physical,
    encode_physical,
    global_checkpoint_path,
    load_autoencoder,
    load_global_autoencoder,
)


RESULT_DIR = (
    ROOT
    / "outputs"
    / "PATH_NAME"
    / "results"
    / "phyloc_inflation_shared"
)
SPATIAL_C2_CONFIG = (
    ROOT
    / "configs"
    / "ks_200cyc.yaml"
)
SPATIAL_C1_CONFIG = (
    ROOT
    / "configs"
    / "ks_200cyc.yaml"
)
GLOBAL_Z256_CONFIG = (
    ROOT
    / "configs"
    / "ks_200cyc.yaml"
)

OUT_BASE = PIC / "Fig08_projection_only_increment_spectra"
OUT_SUMMARY = OUT_BASE.with_name(OUT_BASE.name + "_summary.csv")

DEALIAS_FRAC = 2.0 / 3.0
DEVICE = "cpu"
SPECTRUM_Y_MIN = 1.0e-17
SPECTRUM_Y_SPLIT = 1.0e-7
SPECTRUM_Y_MAX = 1.0e-1
SPECTRUM_LOW_FRACTION = 0.35
SPECTRUM_TICKS = np.array([1.0e-17, 1.0e-15, 1.0e-13, 1.0e-11, 1.0e-9, 1.0e-7, 1.0e-5, 1.0e-3, 1.0e-1])

ACTUAL_METHODS = [
    ("PHY", "N20_test_PHY-EnSRF-loc.npz", "#1F4E79"),
    ("GLA_256", "N20_test_GDLR-EnSRF-loc_global_Z256.npz", "#7B3294"),
    ("SLA_256", "N20_test_SLAT-EnSRF-loc_spatial_C2_Kz128.npz", "#2E7D32"),
    ("SLA_128", "N20_test_SLAT-EnSRF-loc_spatial_C1_Kz128.npz", "#E69F00"),
]

PROJECTION_METHODS = [
    ("PHY", "#1F4E79"),
    ("PHY -> GLA_256", "#7B3294"),
    ("PHY -> SLA_256", "#2E7D32"),
    ("PHY -> SLA_128", "#E69F00"),
]

WINDOWS = [
    ("Cycles 1-50", slice(0, 50)),
    ("Cycles 51-200", slice(50, 200)),
]


def load_yaml(path: Path) -> dict[str, Any]:
    with path.open("r", encoding="utf-8") as f:
        return yaml.safe_load(f)


def one_sided_power(fields: np.ndarray) -> np.ndarray:
    n = fields.shape[-1]
    fft = np.fft.rfft(fields.astype(np.float64), axis=-1)
    power = (np.abs(fft) ** 2) / float(n * n)
    if n > 1:
        power[..., 1:-1] *= 2.0
    return power


def mean_increment_spectrum(increments: np.ndarray, window: slice) -> np.ndarray:
    return one_sided_power(increments[window]).mean(axis=0)


def project_spatial(states: np.ndarray, cfg_path: Path) -> np.ndarray:
    cfg = load_yaml(cfg_path)
    model, mean, std = load_autoencoder(checkpoint_path(cfg), device=DEVICE)
    latent = encode_physical(model, states, mean, std, DEVICE)
    return decode_latent(model, latent, mean, std, DEVICE)


def project_global(states: np.ndarray, cfg_path: Path) -> np.ndarray:
    cfg = load_yaml(cfg_path)
    model, mean, std = load_global_autoencoder(global_checkpoint_path(cfg), device=DEVICE)
    latent = encode_global_physical(model, states, mean, std, DEVICE)
    return decode_global_latent(model, latent, mean, std, DEVICE)


def load_actual_increments() -> tuple[dict[str, np.ndarray], np.ndarray, np.ndarray]:
    actual: dict[str, np.ndarray] = {}
    phy_prior: np.ndarray | None = None
    phy_analysis: np.ndarray | None = None
    for label, filename, _color in ACTUAL_METHODS:
        with np.load(RESULT_DIR / filename) as data:
            prior = np.asarray(data["prior_mean"], dtype=np.float64)
            analysis = np.asarray(data["analysis_mean"], dtype=np.float64)
        actual[label] = analysis - prior
        if label == "PHY":
            phy_prior = prior
            phy_analysis = analysis
    if phy_prior is None or phy_analysis is None:
        raise RuntimeError("Missing PHY prior/analysis arrays.")
    return actual, phy_prior, phy_analysis


def compute_projection_only_increments(phy_prior: np.ndarray, phy_analysis: np.ndarray) -> dict[str, np.ndarray]:
    projected: dict[str, np.ndarray] = {
        "PHY": phy_analysis - phy_prior,
    }
    projected["PHY -> GLA_256"] = project_global(phy_analysis, GLOBAL_Z256_CONFIG) - phy_prior
    projected["PHY -> SLA_256"] = project_spatial(phy_analysis, SPATIAL_C2_CONFIG) - phy_prior
    projected["PHY -> SLA_128"] = project_spatial(phy_analysis, SPATIAL_C1_CONFIG) - phy_prior
    return projected


def band_sum(spec: np.ndarray, lo: int, hi: int) -> float:
    return float(np.nansum(spec[lo : hi + 1]))


def piecewise_log_power_axis(values: np.ndarray | float) -> np.ndarray:
    vals = np.asarray(values, dtype=np.float64)
    clipped = np.clip(vals, SPECTRUM_Y_MIN, SPECTRUM_Y_MAX)
    log_vals = np.log10(clipped)
    log_min = np.log10(SPECTRUM_Y_MIN)
    log_split = np.log10(SPECTRUM_Y_SPLIT)
    log_max = np.log10(SPECTRUM_Y_MAX)

    transformed = np.empty_like(log_vals)
    low_tail = log_vals <= log_split
    transformed[low_tail] = SPECTRUM_LOW_FRACTION * (log_vals[low_tail] - log_min) / (log_split - log_min)
    transformed[~low_tail] = SPECTRUM_LOW_FRACTION + (1.0 - SPECTRUM_LOW_FRACTION) * (
        log_vals[~low_tail] - log_split
    ) / (log_max - log_split)
    return transformed


def spectrum_tick_labels(ticks: np.ndarray) -> list[str]:
    return [rf"$10^{{{int(np.round(np.log10(tick)))}}}$" for tick in ticks]


def main() -> None:
    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    from matplotlib.lines import Line2D

    actual_increments, phy_prior, phy_analysis = load_actual_increments()
    projection_increments = compute_projection_only_increments(phy_prior, phy_analysis)

    n = phy_prior.shape[-1]
    max_mode = int(np.floor((n // 2) * DEALIAS_FRAC))
    modes = np.arange(n // 2 + 1)
    plot_modes = modes[1 : max_mode + 1]

    plt.rcParams.update(
        {
            "font.family": "DejaVu Serif",
            "font.size": 11.0,
            "mathtext.fontset": "dejavuserif",
            "axes.linewidth": 0.9,
        }
    )
    fig, axes = plt.subplots(2, 2, figsize=(11.2, 7.5), sharex=True, sharey=True, facecolor="white")

    rows: list[dict[str, object]] = []
    panel_specs = [
        (axes[0, 0], "actual", "Cycles 1-50", WINDOWS[0][1], actual_increments, ACTUAL_METHODS),
        (axes[0, 1], "projection_only", "Cycles 1-50", WINDOWS[0][1], projection_increments, PROJECTION_METHODS),
        (axes[1, 0], "actual", "Cycles 51-200", WINDOWS[1][1], actual_increments, ACTUAL_METHODS),
        (axes[1, 1], "projection_only", "Cycles 51-200", WINDOWS[1][1], projection_increments, PROJECTION_METHODS),
    ]

    for ax, family, window_label, window_slice, increments_by_label, method_specs in panel_specs:
        for label, *rest in method_specs:
            color = rest[-1]
            spec = mean_increment_spectrum(increments_by_label[label], window_slice)
            ax.semilogx(plot_modes, piecewise_log_power_axis(spec[1 : max_mode + 1]), color=color, lw=1.8)
            rows.append(
                {
                    "panel_family": family,
                    "window": window_label,
                    "method": label,
                    "total_power_k1_682": f"{band_sum(spec, 1, max_mode):.10e}",
                    "low_power_k1_64": f"{band_sum(spec, 1, 64):.10e}",
                    "active_power_k65_256": f"{band_sum(spec, 65, 256):.10e}",
                    "tail_power_k257_682": f"{band_sum(spec, 257, max_mode):.10e}",
                }
            )
        ax.grid(True, which="major", color="0.88", lw=0.8)
        ax.grid(True, which="minor", color="0.93", lw=0.5)
        ax.set_xlim(1, max_mode)
        ax.set_ylim(float(piecewise_log_power_axis(SPECTRUM_Y_MIN)), float(piecewise_log_power_axis(SPECTRUM_Y_MAX)))
        ax.set_axisbelow(True)

    spectrum_ticks = SPECTRUM_TICKS[(SPECTRUM_TICKS >= SPECTRUM_Y_MIN) & (SPECTRUM_TICKS <= SPECTRUM_Y_MAX)]
    spectrum_tick_positions = piecewise_log_power_axis(spectrum_ticks)
    for ax in axes.ravel():
        ax.set_yticks(spectrum_tick_positions)
        ax.set_yticklabels(spectrum_tick_labels(spectrum_ticks))
        ax.axhline(float(piecewise_log_power_axis(SPECTRUM_Y_SPLIT)), color="0.78", lw=0.65, ls=":")

    axes[0, 0].set_title("Actual DA increment")
    axes[0, 1].set_title("PHY analysis projected through AE")
    axes[1, 0].set_xlabel("Wavenumber")
    axes[1, 1].set_xlabel("Wavenumber")
    axes[0, 0].set_ylabel("Cycles 1-50\nincrement power")
    axes[1, 0].set_ylabel("Cycles 51-200\nincrement power")

    tags = ["(a)", "(b)", "(c)", "(d)"]
    for ax, tag in zip(axes.ravel(), tags):
        ax.text(
            0.03,
            0.94,
            tag,
            transform=ax.transAxes,
            ha="left",
            va="top",
            fontsize=13,
            bbox={"facecolor": "white", "edgecolor": "none", "alpha": 0.82, "pad": 1.4},
        )

    actual_handles = [Line2D([0], [0], color=color, lw=2.4, label=label) for label, _filename, color in ACTUAL_METHODS]
    projection_handles = [
        Line2D([0], [0], color=color, lw=2.4, label=label) for label, color in PROJECTION_METHODS
    ]
    leg1 = axes[0, 0].legend(handles=actual_handles, loc="lower left", frameon=False, fontsize=9.2)
    axes[0, 0].add_artist(leg1)
    axes[0, 1].legend(handles=projection_handles, loc="lower left", frameon=False, fontsize=9.2)

    fig.tight_layout()
    fig.savefig(OUT_BASE.with_suffix(".png"), dpi=300)
    fig.savefig(OUT_BASE.with_suffix(".pdf"))
    plt.close(fig)

    with OUT_SUMMARY.open("w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=list(rows[0].keys()))
        writer.writeheader()
        writer.writerows(rows)

    print(OUT_BASE.with_suffix(".png"))
    print(OUT_BASE.with_suffix(".pdf"))
    print(OUT_SUMMARY)


if __name__ == "__main__":
    main()



from pathlib import Path
import os
import sys


SCRIPT_DIR = Path(__file__).resolve().parent
ROOT = SCRIPT_DIR.parent
MPLCONFIGDIR = ROOT / ".mplconfig_pic_fig03_simple"
MPLCONFIGDIR.mkdir(exist_ok=True)
for lock_file in MPLCONFIGDIR.glob("*.matplotlib-lock"):
    try:
        lock_file.unlink()
    except OSError:
        pass
os.environ.setdefault("MPLCONFIGDIR", str(MPLCONFIGDIR))
sys.path.insert(0, str(ROOT))

import matplotlib.pyplot as plt
import numpy as np

from lda.utils import gaspari_cohn


OUTDIR = SCRIPT_DIR
PNG_PATH = OUTDIR / "Fig03_gc_localization_physical_vs_spatial_latent.png"
PDF_PATH = OUTDIR / "Fig03_gc_localization_physical_vs_spatial_latent.pdf"


LATENT_SPACING = 16  # 2048 physical grid points / 128 spatial latent cells
RADII = [16, 32, 48, 64, 96, 128, 192, 256]


def main() -> None:
    plt.rcParams.update(
        {
            "font.family": "DejaVu Serif",
            "font.size": 11,
            "mathtext.fontset": "dejavuserif",
            "axes.linewidth": 0.9,
        }
    )

    fig, axes = plt.subplots(1, 2, figsize=(12.8, 5.2), facecolor="white", sharey=True)
    ax_phys, ax_lat = axes
    distances = np.linspace(0, max(RADII) * 1.08, 900)
    colors = plt.cm.viridis(np.linspace(0.06, 0.96, len(RADII)))

    for radius, color in zip(RADII, colors):
        weights = gaspari_cohn(distances, float(radius))
        ax_phys.plot(distances, weights, color=color, lw=2.4, label=f"R={radius}")

        latent_offsets = np.arange(0, int(np.ceil(radius / LATENT_SPACING)) + 2) * LATENT_SPACING
        latent_offsets = latent_offsets[latent_offsets <= max(RADII) * 1.08]
        latent_weights = gaspari_cohn(latent_offsets, float(radius))
        ax_lat.plot(
            latent_offsets / LATENT_SPACING,
            latent_weights,
            linestyle="-",
            linewidth=2.0,
            alpha=0.65,
            color=color,
        )
        ax_lat.plot(
            latent_offsets / LATENT_SPACING,
            latent_weights,
            linestyle="",
            marker="o",
            markersize=4.6,
            markerfacecolor=color,
            markeredgecolor=color,
            markeredgewidth=1.4,
        )

    ax_phys.set_xlim(0, max(RADII) * 1.06)
    ax_lat.set_xlim(0, max(RADII) / LATENT_SPACING * 1.06)
    ax_phys.set_ylim(-0.035, 1.10)
    ax_phys.set_xlabel("Distance from observation (physical grid points)")
    ax_lat.set_xlabel("Distance from observation (spatial latent cells)")
    ax_phys.set_ylabel("Localization weight")
    for ax in axes:
        ax.grid(True, color="0.88", lw=0.8)
    ax_phys.text(0.02, 0.96, "(a)", transform=ax_phys.transAxes, fontsize=14, ha="left", va="top")
    ax_lat.text(0.02, 0.96, "(b)", transform=ax_lat.transAxes, fontsize=14, ha="left", va="top")
    ax_phys.legend(
        title="Localization radius",
        ncol=2,
        frameon=False,
        loc="upper right",
        columnspacing=1.4,
        handlelength=2.2,
    )

    fig.tight_layout(w_pad=2.0)
    fig.savefig(PNG_PATH, dpi=300, bbox_inches="tight")
    fig.savefig(PDF_PATH, bbox_inches="tight")
    print(f"saved {PNG_PATH}")
    print(f"saved {PDF_PATH}")


if __name__ == "__main__":
    main()
